(** Consistency of the numerical scheme *)

Require Export Fourier.
Require Export one_d_wave_equation.
Open Local Scope nat_scope.
Open Local Scope R_scope.

Section Consistency. 
Variable xmin xmax T:R.
Hypothesis xminLtxmax: xmin < xmax.
Hypothesis T_pos: 0 < T.

Variable partial_derive_firstvar: (R * R -> R) -> (R * R -> R).
Variable partial_derive_secondvar: (R * R -> R) -> (R * R -> R).

Variable c:R.
Variable usol : R * R -> R.
Variable u0 u1: R->R.
Variable f : R*R -> R.

Let u0h := (fun dX j => u0 (fst (Xjk xmin dX j 0))).
Let u1h := (fun dX j => u1 (fst (Xjk xmin dX j 1))).
Let fh  := (fun dX j k => f (Xjk xmin dX j k)).

Hypothesis c_neq_0: c <> 0.
Hypothesis usol_is_sol: is_solution xmin xmax partial_derive_firstvar partial_derive_secondvar c u0 u1 f usol.

Hypothesis f_is_zero_at_beginning: forall x, f (x,0) = 0.

Notation Oupse := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ ni_exact xmin xmax dX)).

Notation ni x := (ni xmin xmax x).
Notation my_proj := (fun X => my_proj xmin xmax X).
Notation scheme := (fun c u0h u1h fh uh dX j n => scheme xmin xmax c u0h u1h fh uh dX j n).
Notation finite_squared_norm_dx := (fun dx u => (finite_squared_norm_dx xmin xmax dx u)).

Lemma OuP_X_to_appr_fst_Betw: forall (P:R*R->Prop)  h g appr,
   (forall x dX, snd (appr (x, 0) dX) = 0) ->
   (forall x dX, P dX -> xmin <= x <= xmax -> xmin <= fst (appr (x,0) dX) <= xmax) -> 
   OuP P (fun (x:{x| Betw xmin xmax x}) dX => h (projT1 x, 0) dX) g ->
   OuP P (fun (X:{x| Betw xmin xmax x}*R) dX => h (appr (projT1 (fst X), 0) dX) dX) g.
unfold OuP; intros P h g appr H H' (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros.
replace (appr (projT1 (fst X), 0) dX) with (fst (appr (projT1 (fst X),0) dX),0).
2: generalize (H (projT1 (fst X)) dX); destruct (appr (projT1 (fst X), 0) dX).
2: simpl; intros H5; now rewrite H5.
assert (Betw xmin xmax (fst (appr (projT1 (fst X), 0) dX))).
apply H'; auto.
destruct X as ((s,L),t); simpl;auto.
replace (fst (appr (projT1 (fst X), 0) dX)) with 
  (projT1 (existT (Betw xmin xmax) (fst (appr (projT1 (fst X), 0) dX)) H5)).
apply H3 with (X:=(existT (Betw xmin xmax) (fst (appr (projT1 (fst X), 0) dX)) H5)); auto.
reflexivity.
Qed.


Lemma OuP_X_to_appr_Betw: forall (P:R*R->Prop) f g appr,
     (forall X dX, P dX -> xmin <= fst X <= xmax -> xmin <= fst (appr X dX) <= xmax) -> 
     OuP P (fun (X:{x| Betw xmin xmax x}*R) dX => f (my_proj X) dX) g -> 
     OuP P (fun (X:{x| Betw xmin xmax x}*R) dX => f (appr (my_proj X) dX) dX) g.
clear; unfold OuP; intros P f g appr H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros.
assert (Betw xmin xmax (fst (appr (my_proj X) dX))).
apply H; auto.
destruct X as ((s,L),t); simpl;auto.
replace (appr (my_proj X) dX) with 
 (my_proj ( (exist (Betw xmin xmax)  (fst (appr (my_proj X) dX)) H5) , snd ((appr (my_proj X) dX)))).
apply H3; auto.
unfold Differential.my_proj; simpl; destruct (appr (projT1 (fst X), snd X) dX); reflexivity.
Qed.


(** Initialization is ok *)

Lemma LL1h_is_second_order: 
   is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3%nat usol ->
   is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4%nat usol ->
  Oups (fun (x:{x| Betw xmin xmax x}) dX => LL1h partial_derive_secondvar c usol ((projT1 x),0) dX) 
           (fun dX => fst dX ^2 + snd dX^2).
intros M; intros.
unfold LL1h.
apply OuP_eq_trans_l with (fun (x:{x| Betw xmin xmax x}) => add2_fun 
  (fun dX => (usol (fst (projT1 x, 0), snd (x, 0) + snd dX) - usol (projT1 x, 0)) / snd dX -
    partial_derive_secondvar usol (projT1 x, 0)
    - c*c*snd dX*/2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol (projT1 x,0))
  (scalar_multiplyA_fun _ (c*c/2)  
     (fun dX  => snd dX * (partial_derive_firstvar_n partial_derive_firstvar 2 usol (projT1 x,0) 
       -(usol (fst (projT1 x, 0) + fst dX, snd (x, 0)) - 2 * usol (projT1 x, 0) +
         usol (fst (projT1 x, 0) - fst dX, snd (x, 0))) / (fst dX * fst dX))))).
intros; unfold add2_fun, scalar_multiplyA_fun.
unfold Rdiv; simpl; ring.
apply OuP_is_EV_plus.
(* DL of order 2 of usol in first variable *)
elim usol_is_sol; unfold L; intros T1 _.
apply OuP_eq_trans_l with
 (fun (a:{x| Betw xmin xmax x}) =>
   (add2_fun (fun dX => (usol (fst (projT1 a, 0), snd (a, 0) + snd dX) - usol (projT1 a, 0)) / snd dX -
     partial_derive_secondvar usol (projT1 a, 0) -
     snd dX * / 2 * partial_derive_secondvar_n partial_derive_secondvar 2 usol (projT1 a, 0)) 
              (fun dX => snd dX * /2 * f (projT1 a,0)))).
intros; unfold add2_fun; rewrite <- T1; try ring.
destruct a; auto with real.
clear T1.
apply OuP_is_EV_plus.
apply OuP_multiply2.
intros dX (K1,K2); auto with real.
apply OuP_X_to_x.
apply OuP_eq_trans_l with 
 (fun (X : {x : R | Betw xmin xmax x} * R) =>
          subtract2_fun (fun dX : R * R => usol (add2 (projT1 (fst X),0)  (0,snd dX)))
            (fun dX : R * R => DL_pol partial_derive_firstvar partial_derive_secondvar (3 - 1) usol (projT1 (fst X),0) (0,snd dX))).
intros a dX (K1,K2).
unfold subtract2_fun, DL_pol, differential, C, partial_derive.
simpl.
replace (add2 (projT1 (fst a), 0) (0, snd dX)) with (projT1 (fst a), 0 + snd dX).
field; auto with real.
unfold add2; simpl; now rewrite Rplus_0_r.
apply OuP_X_to_appr with 
  (appr:= fun X dX => (fst X,0))
  (f:= fun (X : {x : R | Betw xmin xmax x} * R) => 
     subtract2_fun (fun dX : R * R => usol (add2 (my_proj X) (0, snd dX)))
     (fun dX : R * R => DL_pol partial_derive_firstvar partial_derive_secondvar (3 - 1) usol (my_proj X) (0, snd dX))).
apply OuP_inclusion with (fun dX => norm_l2 (0, snd dX) ^ 3).
intros X H1 (H2,H3).
rewrite norm_l2_proj2.
repeat rewrite Rabs_right; try apply Rle_ge; auto with real.
rewrite Rmult_plus_distr_l.
apply Rle_trans with (0+snd X ^ 3); auto with real.
apply Rplus_le_compat.
apply Rmult_le_pos; auto with real.
right; ring.
apply Rmult_le_pos; auto with real.
apply Rle_trans with (0+0); auto with real.
apply Rplus_le_compat; auto with real.
apply OuP_implies_OuP with (P1:=(fun _ => True)); trivial.
apply OuP_snddX_to_dX with 
  (h:= fun (X : {x : R | Betw xmin xmax x} * R) => subtract2_fun (fun dX : R * R => usol (add2 (my_proj X) dX))
     (fun dX : R * R => DL_pol partial_derive_firstvar partial_derive_secondvar (3 - 1) usol (my_proj X) dX))
  (g:=fun dX => norm_l2 dX ^ 3).
trivial.
apply M ; auto.
apply OuP_eq_trans_l with (fun _ _ => 0).
intros; rewrite f_is_zero_at_beginning; ring.
apply OuP_is_EV_zero.
(* discrete partial second order derivative is nearly the continuous one *)
apply OuP_is_EV_mult.
apply Rmult_integral_contrapositive_currified; auto with real.
assert (0 < 2)%R; auto with real.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_divide2.
intros dX (H1,H2); auto with real.
apply OuP_eq_trans_l with 
 (fun (X : {x : R | Betw xmin xmax x}) (dX : R * R) =>
      (fst dX^2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol (projT1 X, 0) -
       (usol (projT1 X + fst dX,  0) - 2 * usol (projT1 X, 0) +
        usol (projT1 X - fst dX,  0)))).
intros a dX (H1,H2); simpl; field; split; auto with real.
apply OuP_X_to_x.
apply OuP_eq_trans_l with
  (scalar_multiplyAA_fun _ _ (-1)
  (fun  (X : {x : R | Betw xmin xmax x} * R) dX  =>
          (usol (add2 (projT1 (fst X),0)  (fst dX, 0)) - 2 * usol (projT1 (fst X),0) + usol (subtract2 (projT1 (fst X),0) (fst dX, 0)) -
           fst dX ^ 2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol (projT1 (fst X),0))%R)).
intros.
unfold scalar_multiplyAA_fun, add2, subtract2; simpl.
rewrite Rplus_0_r; rewrite Rminus_0_r; ring.
apply OuP_multiplyAA.
auto with real.
apply OuP_inclusion with  (fun dX : R * R => fst dX ^ 4).
intros X H1 (H2,H3).
replace (/ snd X * (fst X * (fst X * (fst X ^ 2 + snd X ^ 2)))) with
   ( fst X ^ 4 */ snd X + fst X ^ 2 * snd X).
repeat rewrite Rabs_right.
apply Rle_trans with (fst X ^ 4 * /1 +0)%R;[right; field|idtac].
apply Rplus_le_compat.
apply Rmult_le_compat_l.
auto with real.
apply Rle_Rinv; auto with real.
apply Rle_trans with (2:=H1).
apply Rle_trans with (2:=Rabs_snd_le_norm_l2 X).
rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rmult_le_pos; auto with real.
apply Rle_ge; apply Rle_trans with (0+0); auto with real.
apply Rplus_le_compat; apply Rmult_le_pos; auto with real.
apply Rle_ge; auto with real.
field; auto with real.
apply OuP_X_to_appr with (appr:= fun X dX => (fst X,0))
   (f:= fun (X : {x : R | Betw xmin xmax x} * R) dX => (usol (add2 (my_proj X)  (fst dX, 0)) - 2 * usol (my_proj X) +
    usol (subtract2 (my_proj X)  (fst dX, 0)) -
    fst dX ^ 2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol (my_proj X))%R). 
apply OuP_implies_OuP with (P1:=(fun _ => True)); trivial.
now apply DL_discretization_eps_dx with partial_derive_secondvar.
Qed.   

(** Main lemma: the truncation error is O(dx^2+dt^2) *)
(** Proof by Taylor series and *many* computations *)

Lemma pointwise_consistency:
 is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3 usol ->
 is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4 usol ->
 Oupse (fun (X:({x| Betw xmin xmax x})*R) dX =>
                         truncation_error xmin xmax c usol u0h u1h fh dX (jx (projT1 (fst X)-xmin) (fst dX))  (kt (snd X) (snd dX)))
      (fun dX =>  (fst dX)^2 + (snd dX)^2).
intros HH H.
unfold truncation_error.
apply OuP_eq_trans_l with  
 (fun (X : {x : R | Betw xmin xmax x} * R) dX => match kt (snd X) (snd dX) with
   | 0%nat =>
      scheme c u0h u1h fh (ubar xmin usol) dX (jx (projT1 (fst X) - xmin) (fst dX)) 0%nat
   | S n => match (pred (S n)) with 
      | 0%nat =>
         scheme c u0h u1h fh (ubar xmin usol) dX (jx (projT1 (fst X) - xmin) (fst dX)) 1%nat
      | (S n0) =>
         scheme c u0h u1h fh (ubar xmin usol) dX (jx (projT1 (fst X) - xmin) (fst dX)) (S (S n0))
      end
   end).
intros a dX Y.
case (kt (snd a) (snd dX)); trivial.
intros n; rewrite <- pred_Sn; case n; trivial.
apply OuP_match with (f1:= fun n (X:{x : R | Betw xmin xmax x} * R)  dX => match pred n return R with
       | 0%nat => scheme c u0h u1h fh (ubar xmin usol) dX (jx (projT1 (fst X) - xmin) (fst dX)) 1%nat
       | S n0 => scheme c u0h u1h fh (ubar xmin usol) dX  (jx (projT1 (fst X) - xmin) (fst dX)) (S (S n0))
       end).
(* n = 0 *)
apply OuP_eq_trans_l with (fun _ _ => 0).
intros; unfold one_d_wave_equation.scheme, Lh, ubar.
specialize (jx_Betw xmin xmax (fst a) (fst dX) (proj1 H0)); intros (J1,J2).
unfold u0h.
case (Z_lt_le_dec 0 (jx (projT1 (fst a)-xmin) (fst dX))).
case (Z_lt_le_dec (jx (projT1 (fst a)-xmin) (fst dX)) (ni (fst dX))); intros.
unfold discretization.
destruct usol_is_sol as (_,(_,(H1,_))).
unfold L0 in H1; rewrite <- H1.
unfold Xjk; simpl.
rewrite Rmult_0_l; ring.
apply Xjk_Betw.
apply H0.
replace (jx (projT1 (fst a) - xmin) (fst dX)) with (ni (fst dX)) by auto with zarith.
rewrite zeroness_ubar_2 with (1:=usol_is_sol); trivial.
destruct H0; auto with real.
apply H0.
intros; replace (jx (projT1 (fst a) - xmin) (fst dX)) with 0%Z by auto with zarith.
rewrite zeroness_ubar_1 with (1:=usol_is_sol); trivial.
apply OuP_is_EV_zero.
apply OuP_match with (f1 := fun n (X:{x : R | Betw xmin xmax x} * R) dX =>
     scheme c u0h u1h fh (ubar xmin usol) dX (jx (projT1 (fst X) - xmin) (fst dX)) (S n)).
(* n = 1 *)
apply OuP_case with (f1:= fun (X:{x : R | Betw xmin xmax x} * R) dX => LL1h partial_derive_secondvar c usol (Xxt xmin (projT1 (fst X),0) dX) dX)
  (f2 := fun _ _ => 0).
intros.
specialize (jx_Betw xmin xmax (fst X) (fst dX) (proj1 H0)); intros (J1,J2).
case (Zle_lt_or_eq _ _ J1).
case (Zle_lt_or_eq _ _ J2).
clear J1 J2; intros J2 J1.
cut (jx (projT1 (fst X) - xmin) (fst dX) <= ni (fst dX) -1)%Z; [clear J2; intros J2 |auto with zarith].
case (Zle_lt_or_eq _ _ J2).
(* + 1 <= j < ni-1 *)
clear J2; intros J2.
left.
unfold one_d_wave_equation.scheme, Lh, Ah, LL1h, Xxt.
simpl.
unfold u1h.
destruct usol_is_sol as (_,(H4,_)).
unfold L1 in H4; rewrite <- H4.
case (Z_lt_le_dec 0 (jx (projT1 (fst X) - xmin) (fst dX))); 
  [intros _|intros J3; contradict J1; auto with zarith].
case (Z_lt_le_dec (jx (projT1 (fst X) - xmin) (fst dX)) (ni (fst dX)));
  [intros _|intros J3; contradict J2; auto with zarith].
repeat rewrite ubar0_ubar with (1:=usol_is_sol); try apply H0; auto with zarith.
unfold ubar,discretization, Xjk.
simpl.
replace (INR (kt 0 (snd dX))) with 0%R.
rewrite Rmult_0_l.
rewrite Rplus_0_l.
rewrite Rmult_1_l.
replace (xmin + IZR (jx ((projT1 (fst X)-xmin)) (fst dX) + 1) * fst dX) with 
(xmin + IZR (jx (projT1 (fst X)-xmin) (fst dX)) * fst dX + fst dX).
replace (xmin + IZR (jx (projT1 (fst X)-xmin) (fst dX) - 1) * fst dX) with
  (xmin + IZR (jx (projT1 (fst X)-xmin) (fst dX)) * fst dX - fst dX).
field.
destruct H0 as (Y1,(Y2,Y3)); split; auto with real.
unfold Zminus; rewrite plus_IZR; simpl; ring.
rewrite plus_IZR; simpl; ring.
unfold kt.
replace (0/snd dX) with (IZR 0).
rewrite Int_part_IZR.
now simpl.
unfold Rdiv; simpl; ring.
apply Xjk_Betw; apply H0.
(* + j = ni-1 *)
intros J3; rewrite J3.
left.
rewrite scheme_eq_3;[idtac|auto with zarith].
unfold Lh, LL1h, Xxt, Xjk, Ah; simpl.
rewrite J3.
replace (INR (kt 0 (snd dX))) with 0%R.
rewrite Rmult_0_l.
rewrite Rplus_0_l.
replace (ni (fst dX) - 1 + 1)%Z with (ni (fst dX)) by ring.
unfold u1h.
destruct usol_is_sol as (_,(H4,_)).
unfold L1 in H4; rewrite <- H4.
unfold ubar, discretization, Xjk.
simpl; rewrite Rmult_0_l, Rmult_1_l.
replace (xmin + IZR (ni (fst dX) - 1 - 1) * fst dX) with (xmin + IZR (ni (fst dX) - 1) * fst dX - fst dX).
replace (xmin + IZR (ni (fst dX) - 1) * fst dX + fst dX) with (xmin + IZR (ni (fst dX)) * fst dX).
rewrite zeroness_usol_xmin_ni with (1:=usol_is_sol).
field.
destruct H0 as (Y1,(Y2,Y3)); split; auto with real.
destruct H0 as (Y1,_); auto with real.
apply H0.
unfold Zminus; rewrite plus_IZR; simpl; ring.
unfold Zminus; repeat rewrite plus_IZR; simpl; ring.
apply Xjk_Betw_j.
apply H0.
auto with zarith.
unfold kt.
replace (0/snd dX) with (IZR 0).
rewrite Int_part_IZR.
now simpl.
unfold Rdiv; simpl; ring.
(* + j=ni *)
intros J3 J4; rewrite J3.
right.
rewrite scheme_eq_2;[idtac|auto with zarith].
apply zeroness_ubar_2 with (1:=usol_is_sol).
destruct H0; auto with real.
apply H0.
(* + j=0 *)
intros J3; rewrite <- J3.
right.
rewrite scheme_eq_1;[idtac|auto with zarith].
apply zeroness_ubar_1 with (1:=usol_is_sol).
(* + LL1h *)
apply OuP_eq_trans_l with (fun (X : {x : R | Betw xmin xmax x} * R) (dX : R * R) =>
   LL1h partial_derive_secondvar c usol (Xxt  xmin (my_proj ((fst X), 0)) dX) dX).
intros; reflexivity.
apply OuP_X_to_appr_fst_Betw.
intros; unfold Xxt, kt; simpl.
replace (0 / snd dX) with (IZR 0).
rewrite Int_part_IZR.
simpl.
now rewrite Rmult_0_l.
unfold Rdiv; simpl; ring.
unfold Xxt, Xjk,jx; simpl.
clear; intros x dX (H1,H2) (H3,H4); split.
apply Rle_trans with (xmin+0);[auto with real|apply Rplus_le_compat_l].
apply Rmult_le_pos; auto with real.
replace 0%R with (IZR 0) by reflexivity.
apply IZR_le.
apply Int_part_pos.
unfold Rdiv; apply Rmult_le_pos; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify; assumption.
apply Rplus_le_reg_l with (-xmin).
apply Rmult_le_reg_l with (/fst dX); auto with real.
apply Rle_trans with (IZR (Int_part ((x - xmin) / fst dX))).
right; field; auto with real.
apply Rle_trans with ((x - xmin) / fst dX).
apply (proj1 (base_Int_part _)).
unfold Rdiv; rewrite Rmult_comm.
apply Rmult_le_compat_l; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify; assumption.
apply OuP_implies_OuP with (fun dX => 0 < fst dX /\ 0 < snd dX).
intros dX (Y1,(Y2,Y3)); split; assumption.
now apply LL1h_is_second_order.
apply OuP_is_EV_zero.
(* autres n *)
apply OuP_case with (f1:= fun (X:{x : R | Betw xmin xmax x} * R) dX => 
 ((ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX))
       (S (S (pred (pred (kt (snd X) (snd dX)))))) -
     2 *
     ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX))
       (S (pred (pred (kt (snd X) (snd dX))))) +
     ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX))
       (pred (pred (kt (snd X) (snd dX))))) /
    (snd dX * snd dX) +
    Ah c
      (fun j' : Z =>
       ubar xmin usol dX j' (S (pred (pred (kt (snd X) (snd dX))))))
      (fst dX) (jx (projT1 (fst X)-xmin) (fst dX)) -
    fh dX (jx (projT1 (fst X)-xmin) (fst dX))
      (S (pred (pred (kt (snd X) (snd dX)))))))
   (f2 := fun _ _ => 0).
intros X dX (Y1,(Y2,Y3)).
specialize (jx_Betw xmin xmax (fst X) (fst dX) Y1); intros (J1,J2).
case (Zle_lt_or_eq _ _ J1).
case (Zle_lt_or_eq _ _ J2).
clear J1 J2; intros J2 J1.
(* + 1 <= j < ni-1 *)
left.
rewrite scheme_eq_3; auto with zarith.
unfold Lh.
repeat rewrite <- pred_Sn.
replace (S (S (pred (pred (kt (snd X) (snd dX)))))) with
   (pred (pred (kt (snd X) (snd dX))) + 2)%nat by omega.
replace (S (pred (pred (kt (snd X) (snd dX))))) with 
   (pred (pred (kt (snd X) (snd dX))) + 1)%nat by omega.
reflexivity.
(* + j = ni *)
intros J3 J4; rewrite J3.
right.
rewrite scheme_eq_2; auto with zarith.
apply zeroness_ubar_2 with (1:=usol_is_sol); auto with real.
(* + j=0 *)
intros J3; rewrite <- J3.
right.
rewrite scheme_eq_1; auto with zarith.
apply zeroness_ubar_1 with (1:=usol_is_sol); auto with real.
(* + end of cases *)
Notation blop :=  (fun X dX => (S (pred (pred (kt (snd X) (snd dX)))))).
Notation Xxtd :=  (fun (X : {x : R | Betw xmin xmax x} * R) dX => (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (S (pred (pred (kt (snd X) (snd dX))))))).
apply OuP_eq_trans_l with (fun (X : {x : R | Betw xmin xmax x} * R) => add2_fun 
   (fun dX => (((ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX)) ((blop X dX) + 1)%nat -
      2 * ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX) +
      ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX - 1)) / (snd dX * snd dX))
         -  partial_derive_secondvar_n partial_derive_secondvar 2 usol (Xxtd X dX)))
   (fun dX => ((Ah c (fun j : Z => ubar xmin usol dX j (blop X dX)) (fst dX) (jx (projT1 (fst X)-xmin) (fst dX)))
          + c ^ 2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol (Xxtd X dX)))).
intros; destruct a as (r,r0); destruct dX as (r1,r2); unfold add2_fun.
replace (partial_derive_secondvar_n partial_derive_secondvar 2 usol (Xxtd (r, r0) (r1, r2))) with
  (c ^ 2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol (Xxtd (r, r0) (r1, r2))
    + f (Xxtd (r, r0) (r1, r2))).
simpl; unfold fh.
replace (S (pred (pred (kt r0 r2)) + 1)) with (S (S (pred (pred (kt r0 r2))))) by omega.
replace (pred (pred (kt r0 r2)) - 0)%nat with (pred (pred (kt r0 r2))) by omega.
ring.
destruct usol_is_sol as (T1,(T2,T3)); unfold L in T1.
rewrite <- T1; try ring.
apply Xjk_Betw; apply H0.
apply OuP_is_EV_plus.
(* derivee en temps *)
apply OuP_inclusion with (fun dX : R * R => snd dX ^ 2).
intros.
unfold pow; repeat rewrite Rmult_1_r.
fold (Rsqr (snd X)); fold (Rsqr (fst X)).
repeat rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rle_trans with (0 + Rsqr (snd X) ); auto with real.
apply Rle_trans with (0 + 0); auto with real.
apply OuP_multiply2.
intros dX (H1,(H2,H3)); auto with real.
apply OuP_multiply2.
intros dX (H1,(H2,H3)); auto with real.
apply OuP_eq_trans_l with 
 (fun (X : {x : R | Betw xmin xmax x} * R) dX =>
     ((ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX)) ((blop X dX) + 1) -
       2 * ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX) +
       ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX - 1)) 
     - snd dX^2 * partial_derive_secondvar_n partial_derive_secondvar 2 
        usol (Xxtd X dX))).
intros a dX (H1,(H2,H3)); field; auto with real.
apply OuP_eq_trans_r with (fun dX : R * R => snd dX ^ 4).
intros; ring.
unfold ubar, discretization, Xxt.
apply OuP_eq_trans_l with 
  (fun (X : {x : R | Betw xmin xmax x} * R) dX =>
   (usol (add2 (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))) (0, snd dX)) -
    2 * usol (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))) +
    usol (subtract2 (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))) (0, snd dX)) -
    snd dX ^ 2 * partial_derive_secondvar_n partial_derive_secondvar 2 usol
      (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))))).
intros.
replace (add2 (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX)) (blop a dX)) (0, snd dX)) with
   (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX)) (blop a dX + 1)).
replace (subtract2 (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX)) (blop a dX)) (0, snd dX)) with
   (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX)) (blop a dX - 1)).
ring.
unfold subtract2, Xjk.
replace (S (pred (pred (kt (snd a) (snd dX)))) - 1)%nat with
   (pred (pred (kt (snd a) (snd dX)))) by omega.
simpl.
replace (xmin + IZR (jx (projT1 (fst a)-xmin) (fst dX)) * fst dX - 0) with (xmin +IZR (jx (projT1 (fst a)-xmin) (fst dX)) * fst dX) by ring.
case (pred (pred (kt (snd a) (snd dX)))).
replace (INR 0 * snd dX) with (1 * snd dX - snd dX) by (simpl; ring); auto.
intros.
replace (INR (S n) * snd dX) with ((INR (S n) + 1) * snd dX - snd dX) by ring; auto.
unfold add2, Xjk.
replace (snd
  (xmin + IZR (jx (projT1 (fst a)-xmin) (fst dX)) * fst dX,
  INR (S (pred (pred (kt (snd a) (snd dX))))) * snd dX) + snd (0, snd dX)) with
  (INR (S (pred (pred (kt (snd a) (snd dX)))) + 1) * snd dX).
simpl.
replace (xmin+IZR (jx (projT1 (fst a)-xmin) (fst dX)) * fst dX + 0) with (xmin+IZR (jx (projT1 (fst a)-xmin) (fst dX)) * fst dX) by ring; auto.
rewrite plus_INR; simpl; ring.
apply OuP_X_to_appr_Betw with 
  (appr:= (fun X dX 
           => (Xjk xmin dX (jx ((fst X)-xmin) (fst dX)) (S (pred (pred (kt (snd X) (snd dX))))))))
  (f := fun X dX => usol (add2 X (0, snd dX)) -2 *usol X + usol (subtract2 X (0, snd dX))
     -  snd dX ^ 2 * partial_derive_secondvar_n partial_derive_secondvar 2 usol X).
clear; unfold Xjk, jx; simpl; intros x dX (H1,H2) (H3,H4); split.
apply Rle_trans with (xmin+0);[auto with real|apply Rplus_le_compat_l].
apply Rmult_le_pos; auto with real.
replace 0%R with (IZR 0) by reflexivity.
apply IZR_le.
apply Int_part_pos.
unfold Rdiv; apply Rmult_le_pos; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify; assumption.
apply Rplus_le_reg_l with (-xmin).
apply Rmult_le_reg_l with (/fst dX); auto with real.
apply Rle_trans with (IZR (Int_part ((fst x - xmin) / fst dX))).
right; field; auto with real.
apply Rle_trans with ((fst x - xmin) / fst dX).
apply (proj1 (base_Int_part _)).
unfold Rdiv; rewrite Rmult_comm.
apply Rmult_le_compat_l; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify; assumption.
apply OuP_implies_OuP with (P1:=fun _ => True).
intros; auto.
now apply DL_discretization_eps_dt with (partial_derive_firstvar:=partial_derive_firstvar)
   (partial_derive_secondvar:=partial_derive_secondvar) (f:=usol).
(* derivee en espace *)
apply OuP_inclusion with (fun dX : R * R => fst dX ^ 2).
intros.
unfold pow; repeat rewrite Rmult_1_r.
fold (Rsqr (snd X)); fold (Rsqr (fst X)).
repeat rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rle_trans with (Rsqr (fst X)+0); auto with real.
apply Rle_trans with (0 + 0); auto with real.
unfold Ah.
apply OuP_eq_trans_l with (scalar_multiplyAA_fun _ _ (-c*c) 
   (fun (a : {x : R | Betw xmin xmax x} * R)  dX => (ubar xmin usol dX (jx (projT1 (fst a)-xmin) (fst dX) + 1) (blop a dX) -
    2 * ubar xmin usol dX (jx (projT1 (fst a)-xmin) (fst dX)) (blop a dX) +
    ubar xmin usol dX (jx (projT1 (fst a)-xmin) (fst dX) - 1) (blop a dX)) /
   (fst dX * fst dX) 
  - partial_derive_firstvar_n partial_derive_firstvar 2 usol (Xxtd a dX))).
intros; unfold scalar_multiplyAA_fun; simpl.
unfold Rdiv; ring.
apply OuP_is_EV_mult.
auto with real.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_eq_trans_l with 
 (fun   (X : {x : R | Betw xmin xmax x} * R) dX  =>
     ((ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX) + 1) (blop X dX) -
       2 * ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX) +
       ubar xmin usol dX (jx (projT1 (fst X)-xmin) (fst dX) - 1) (blop X dX)) 
     - fst dX^2 * partial_derive_firstvar_n partial_derive_firstvar 2 
        usol (Xxtd X dX))).
intros a dX (H1,H2); field; auto with real.
apply OuP_eq_trans_r with (fun dX : R * R => fst dX ^ 4).
intros; ring.
unfold ubar, discretization, Xxt.
apply OuP_eq_trans_l with 
  (fun   (X : {x : R | Betw xmin xmax x} * R) dX =>
   (usol (add2 (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX)) (fst dX,0)) -
    2 * usol (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX)) +
    usol (subtract2 (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX)) (fst dX,0)) -
    fst dX ^ 2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol
      (Xjk xmin dX (jx (projT1 (fst X)-xmin) (fst dX)) (blop X dX)))).
intros.
replace (add2 (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX)) (blop a dX)) (fst dX, 0)) with
   (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX) + 1) (blop a dX)).
replace (subtract2 (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX)) (blop a dX)) (fst dX, 0)) with
   (Xjk xmin dX (jx (projT1 (fst a)-xmin) (fst dX) - 1) (blop a dX)).
ring.
unfold add2, subtract2, Xjk; simpl.
unfold Zminus; rewrite plus_IZR; simpl.
replace (xmin+(IZR (jx (projT1 (fst a)-xmin) (fst dX)) + -1) * fst dX) with 
  (xmin+IZR (jx (projT1 (fst a)-xmin) (fst dX)) * fst dX + - fst dX) by ring; auto with real.
unfold add2, Xjk; simpl.
rewrite plus_IZR; simpl.
replace (xmin+(IZR (jx (projT1 (fst a)-xmin) (fst dX)) + 1) * fst dX) with 
  (xmin+IZR (jx (projT1 (fst a)-xmin) (fst dX)) * fst dX + fst dX) by ring; auto with real.
apply OuP_X_to_appr_Betw with 
  (appr:= fun X dX => (Xjk xmin dX (jx (fst X-xmin) (fst dX)) (blop X dX)))
(f := fun X dX => usol (add2 X (fst dX, 0)) -2 *usol X + usol (subtract2 X (fst dX, 0))
     -  fst dX ^ 2 * partial_derive_firstvar_n partial_derive_firstvar 2 usol X).
clear; unfold Xjk, jx; simpl; intros x dX (H1,H2) (H3,H4); split.
apply Rle_trans with (xmin+0);[auto with real|apply Rplus_le_compat_l].
apply Rmult_le_pos; auto with real.
replace 0%R with (IZR 0) by reflexivity.
apply IZR_le.
apply Int_part_pos.
unfold Rdiv; apply Rmult_le_pos; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify; assumption.
apply Rplus_le_reg_l with (-xmin).
apply Rmult_le_reg_l with (/fst dX); auto with real.
apply Rle_trans with (IZR (Int_part ((fst x - xmin) / fst dX))).
right; field; auto with real.
apply Rle_trans with ((fst x - xmin) / fst dX).
apply (proj1 (base_Int_part _)).
unfold Rdiv; rewrite Rmult_comm.
apply Rmult_le_compat_l; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify; assumption.
apply OuP_implies_OuP with (P1:=fun _ => True).
intros; auto.
now apply DL_discretization_eps_dx with (partial_derive_firstvar:=partial_derive_firstvar)
   (partial_derive_secondvar:=partial_derive_secondvar) (f:=usol).
apply OuP_is_EV_zero. 
Qed.

(** From one point to an infinite squared norm *)

Lemma Oup_pointwise_to_squared_norm_dx : forall vh, forall g,
  Oupse
    (fun (X:({x| Betw xmin xmax x})*R) dX => vh dX (jx (projT1 (fst X)-xmin) (fst dX)) (kt (snd X) (snd dX)))
    g ->
  Oupse
    (fun (t : {t|Under T t}) dX =>
      finite_squared_norm_dx (fst dX) (fun j => vh dX j (kt (projT1 t) (snd dX))))
    (fun dX => Rsqr (g dX)).
Proof.
intros.
destruct H as (alp, (C, (H1,(H2,H3)))); intros.
exists alp; exists (C * C * (xmax-xmin)).
split; trivial.
split.
repeat apply Rmult_lt_0_compat; trivial.
apply Rplus_lt_reg_r with xmin; ring_simplify.
apply xminLtxmax.
intros t (dx, dt) H4; simpl; intros (Y1, (Y2,Y3)).
rewrite Rabs_right.
apply Rle_trans with
  ((xmax - xmin)*Rsqr (C*(g (dx,dt)))).
apply finite_squared_norm_dx_majoration; auto with real.
intros.
rewrite Rsqr_abs; rewrite (Rsqr_abs (C * g (dx, dt))).
apply Rsqr_incr_1; try apply Rabs_pos.
assert (Betw xmin xmax (xmin+IZR j * dx)).
unfold Betw; split.
apply Rle_trans with (xmin+IZR 0 * dx)%R.
right; simpl; ring.
apply Rplus_le_compat_l; apply Rmult_le_compat_r; auto with real.
apply IZR_le; apply H.
apply Rle_trans with (xmin+IZR ((ni dx)) * dx)%R.
apply Rplus_le_compat_l; apply Rmult_le_compat_r; auto with real.
apply IZR_le.
destruct H; omega.
unfold ni_exact in Y3; simpl in Y3; rewrite Y3.
right; field; auto with real.
replace j with (jx (projT1 (exist (Betw xmin xmax) (xmin + IZR j * dx) H0)-xmin) dx).
rewrite Rabs_mult.
rewrite (Rabs_right C).
apply H3 with (dX := (dx, dt)) (X:=((exist (Betw xmin xmax) (xmin + IZR j * dx) H0),projT1 t)); auto.
apply Rle_ge; auto with real.
unfold jx; simpl.
ring_simplify (((xmin + IZR j * dx - xmin))).
replace (IZR j * dx / dx) with (IZR j).
apply Int_part_IZR.
field; auto with real.
apply Rle_trans with ((xmax - xmin) * (C * g (dx, dt))²)%R.
apply Rmult_le_compat_r; auto with real.
rewrite Rabs_right; auto with real.
unfold Rsqr; right; ring.
apply Rle_ge; apply finite_squared_norm_dx_pos; auto with real.
Qed.




Lemma Oup_pointwise_to_squared_norm_dx_truncation_error: forall g,
  Oupse
    (fun (X:({x| Betw xmin xmax x})*R) dX =>
      truncation_error xmin xmax c usol u0h u1h fh dX (jx ((projT1 (fst X)-xmin)) (fst dX)) (kt (snd X) (snd dX))) 
    g ->
  Oupse
    (fun (t : {t|Under T t}) dX =>
      finite_squared_norm_dx (fst dX)
        (fun j => truncation_error xmin xmax c usol u0h u1h fh dX j (kt (projT1 t) (snd dX))))
    (fun dX => Rsqr (g dX)).
Proof.
intros.
apply (Oup_pointwise_to_squared_norm_dx (truncation_error xmin xmax c usol u0h u1h fh)).
exact H.
Qed.

Theorem consistency: 
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3%nat usol -> 
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4%nat usol -> 
  scheme_is_consistent_of_order xmin xmax T c usol u0h u1h fh 2%nat 2%nat.
Proof.
intros; unfold scheme_is_consistent_of_order; intros.
apply
  (Oup_pointwise_to_squared_norm_dx_truncation_error
    (fun dX : R * R => (fst dX) ^ 2 + (snd dX) ^ 2)); auto.
apply pointwise_consistency; auto.
Qed.

End Consistency. 
