(** Convergence of the numerical scheme *)

Require Export stability.
Require Export consistency.
Open Local Scope nat_scope.
Open Local Scope R_scope.

Section Conv.
Variable xmin xmax T:R.
Hypothesis xminLtxmax: xmin < xmax.
Hypothesis T_pos: 0 < T.

Variable partial_derive_firstvar: (R * R -> R) -> (R * R -> R).
Variable partial_derive_secondvar: (R * R -> R) -> (R * R -> R).

Variable c:R.
Variable usol : R * R -> R.
Variable u0 u1: R->R.
Variable f : R*R -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable eps:R.

Let u0h := (fun dX j => u0 (fst (Xjk xmin dX j 0))).
Let u1h := (fun dX j => u1 (fst (Xjk xmin dX j 1))).
Let fh  := (fun dX j k => f (Xjk xmin dX j k)).


(** All the hypotheses *)

Hypothesis usol_is_sol: is_solution xmin xmax partial_derive_firstvar partial_derive_secondvar c u0 u1 f usol.
Hypothesis unh_is_discrete_sol: forall dX, is_discrete_solution xmin xmax c u0h u1h fh unh dX.

Hypothesis f_is_zero_at_beginning: forall x, f (x,0) = 0.

Hypothesis c_pos: 0 < c.
Hypothesis eps_lt: 0 < eps < 1.


Notation Oupse := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ ni_exact xmin xmax dX)).
Notation OupsCFLe := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ CFL_condition_r c eps dX /\ ni_exact xmin xmax dX)).

Notation ni x := (ni xmin xmax x).
Notation finite_squared_norm_dx := (fun dx u => (finite_squared_norm_dx xmin xmax dx u)).
Notation finite_norm_dx := (fun dx u => (finite_norm_dx xmin xmax dx u)).
Notation energy := (fun c uh dX n => energy xmin xmax c uh dX n).


Lemma unh_0: forall dX j, 0 < fst dX -> snd dX <> 0 -> ni_exact xmin xmax dX -> 
   (0 <= j <= ni (fst dX))%Z -> unh dX j 0 = ubar xmin usol dX j 0.
intros.
generalize  (unh_is_discrete_sol dX H0 j 0%nat H2).
unfold scheme,u0h.
case (Z_lt_le_dec 0 j).
case (Z_lt_le_dec j (ni (fst dX))).
intros J1 J2.
replace (Lh c (unh dX) dX j 0) with (unh dX j 0).
replace (u0 (fst (Xjk xmin dX j 0))) with (ubar xmin usol dX j 0);auto with real.
elim usol_is_sol;unfold L0; intros _ (_,(H3,_)).
unfold ubar,discretization, Xjk.
rewrite Rmult_0_l;apply H3.
replace (fst (xmin + IZR j * fst dX, 0)) with  (fst (Xjk xmin dX j 0)) by reflexivity.
apply Xjk_Betw_j; auto with zarith.
now unfold Lh.
intros J1 J2; replace j with (ni (fst dX)); auto with zarith.
intros Y; rewrite Y, zeroness_ubar_2 with (1:=usol_is_sol); auto with real.
intros J1; replace j with 0%Z; auto with zarith.
intros Y; rewrite Y, zeroness_ubar_1 with (1:=usol_is_sol); auto.
Save.

(** The convergence error is solution of the same scheme with specific inputs *)

Lemma convergence_error_is_discrete_sol: forall dX j n, 
  0 < fst dX -> 0 < snd dX -> ni_exact xmin xmax dX ->
 (0 <= j <= ni (fst dX))%Z ->
  scheme xmin xmax c
      (fun dX j => 0) 
      (fun dX j => (convergence_error xmin usol unh dX j 1) / (snd dX)) 
      (fun dX j m => truncation_error xmin xmax c usol u0h u1h fh dX j (S m))
      (convergence_error xmin usol unh) dX j n = 0.
unfold is_discrete_solution, scheme in unh_is_discrete_sol.
unfold scheme, convergence_error; intros.
assert (M: snd dX <> 0) by auto with real.
specialize (unh_is_discrete_sol dX M j n); simpl in unh_is_discrete_sol.
generalize unh_is_discrete_sol.
case (Z_lt_le_dec 0 j); intros J1.
case (Z_lt_le_dec j (ni (fst dX))); intros J2.
case n.
(* *)
rewrite Lh_is_linear.
intros H3.
apply trans_eq with ((Lh c (ubar xmin usol dX) dX j 0 - u0h dX j) -
   (Lh c (unh dX) dX j 0 - u0h dX j));[ring|unfold Lh].
rewrite <- unh_0; auto with real zarith.
(* *)
intros n0; case n0.
rewrite Lh_is_linear.
simpl.
rewrite unh_0; auto with real zarith.
rewrite Ah_eq with xmin xmax c (fun j' => unh dX j' 0) (fun j' => ubar xmin usol dX j' 0) (fst dX) j.
intros; field; auto with real.
intros j0; apply unh_0; auto with zarith real.
auto with zarith.
(* *)
intros.
rewrite Lh_is_linear.
apply trans_eq with ((-(Lh c (unh dX) dX j (S (S n1)) - fh dX j (S n1)) +
   (-fh dX j (S n1)+ Lh c (ubar xmin usol dX) dX j (S (S n1)) 
   - truncation_error xmin xmax c usol u0h u1h fh dX j (S (S n1)))));[ring|idtac].
unfold Lh.
rewrite unh_is_discrete_sol0.
unfold truncation_error, scheme.
case (Z_lt_le_dec 0 j);[intros _| intros J3; contradict J1; auto with zarith].
case (Z_lt_le_dec j (ni (fst dX))); [intros _ | intros J3; contradict J2; auto with zarith].
unfold Lh; ring.
split; auto with zarith.
(* *)
intros Y; rewrite Y; auto.
replace j with (ni (fst dX)) by auto with zarith.
rewrite zeroness_ubar_2 with (1:=usol_is_sol); auto with real.
(* *)
intros Y; rewrite Y; auto.
replace j with 0%Z by auto with zarith.
rewrite zeroness_ubar_1 with (1:=usol_is_sol); auto with real.
Qed.

(** Initializations are ok *)

Lemma condinite1j_aux: forall ph qh dX,
  (snd dX <> 0) ->
  (forall j, Lh c ph dX j 1 = Lh c qh dX j 1) ->
  (forall j, ph j 0%nat = qh j 0%nat) ->
  (forall j, ph j 1%nat = qh j 1%nat).
clear; intros.
specialize (H0 j).
unfold Lh in H0.
apply Rmult_eq_reg_l with (/snd dX); auto with real.
apply Rplus_eq_reg_l with  ((-ph j 0%nat) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => ph j' 0%nat) (fst dX) j).
apply trans_eq with ((ph j 1%nat - ph j 0%nat) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => ph j' 0%nat) (fst dX) j).
unfold Rdiv; ring.
rewrite H0.
unfold Ah; repeat rewrite H1.
unfold Rdiv; ring.
Qed.


Lemma e1h_LL1h: forall x dX, 0 < fst dX -> snd dX<>0-> ni_exact xmin xmax dX ->
  (0 < jx (x-xmin) (fst dX) < ni (fst dX))%Z ->
  convergence_error xmin usol unh dX (jx (x-xmin) (fst dX)) 1 / (snd dX) =
  LL1h partial_derive_secondvar c usol (Xxt xmin (x,0) dX) dX. 
intros x dX H H0 M K; unfold convergence_error,LL1h.
elim usol_is_sol.
unfold L1; intros H1 (H2,(H3,H4)).
unfold Xxt; simpl.
replace (kt 0 (snd dX)) with 0%nat.
unfold Xjk; rewrite Rmult_0_l.
rewrite H2.
set (j := jx (x-xmin) (fst dX)).
replace (u1 (xmin+IZR j * fst dX)) with 
   (Lh c (unh dX) dX j 1).
unfold Lh,Ah.
repeat rewrite unh_0; auto with real.
unfold ubar,discretization.
unfold Xjk.
rewrite Rmult_0_l.
replace (xmin+IZR j * fst dX + fst dX) with (xmin+IZR (j + 1) * fst dX).
replace (xmin+IZR j * fst dX - fst dX) with (xmin+IZR (j - 1) * fst dX).
rewrite Rmult_1_l,Rplus_0_l.
field.
split; auto with real.
unfold Zminus.
rewrite plus_IZR, Ropp_Ropp_IZR.
simpl;ring.
rewrite plus_IZR;simpl;ring.
unfold j; auto with zarith.
unfold j; auto with zarith.
unfold j; auto with zarith.
generalize (unh_is_discrete_sol dX H0 j 1%nat).
unfold scheme.
case (Z_lt_le_dec 0 j);[intros _|intros J; contradict J; unfold j; auto with zarith].
case (Z_lt_le_dec j (ni (fst dX)));[intros _|intros J; contradict J; unfold j; auto with zarith].
replace (u1h dX j) with (u1 (xmin+IZR j * fst dX)).
assert ((0 <= j <= ni (fst dX))%Z); unfold j; auto with zarith.
intros Y; specialize (Y H5); auto with real.
now unfold u1h.
replace (xmin + IZR (jx (x - xmin) (fst dX)) * fst dX) with 
  (fst (Xjk xmin dX (jx (x - xmin) (fst dX)) 0)) by reflexivity. 
apply Xjk_Betw_j; auto with real zarith.
unfold kt.
replace (0 / snd dX) with (IZR 0).
now rewrite Int_part_IZR.
unfold Rdiv.
now rewrite Rmult_0_l.
Qed.


(* lemme 72-210709*)
Lemma condinite1j: 
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3%nat usol ->
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4%nat usol ->
  Oupse  
   (fun (x : {x : R | Betw xmin xmax x}) dX => convergence_error xmin usol unh dX (jx  (projT1 x-xmin) (fst dX)) 1 / (snd dX))
   (fun dX => fst dX ^2 + snd dX^2).
intros M; intros.
apply OuP_case with (f1:= fun (x : {x : R | Betw xmin xmax x}) dX => 0) 
   (f2:=fun (x : {x : R | Betw xmin xmax x}) dX => LL1h partial_derive_secondvar c usol (Xxt xmin (projT1 x,0) dX) dX).
intros x dX (H1,(H2,H3)).
case (Z_lt_le_dec 0 (jx (projT1 x-xmin) (fst dX))).
intros.
case (Z_lt_le_dec (jx (projT1 x-xmin) (fst dX)) (ni (fst dX))).
intros; right.
apply e1h_LL1h; auto with real.
intros; left.
unfold convergence_error.
replace (jx (projT1 x - xmin) (fst dX)) with (ni (fst dX)).
rewrite zeroness_ubar_2 with (1:=usol_is_sol); auto with real.
rewrite zeroness_unh_2 with (2:=unh_is_discrete_sol dX); auto with real.
unfold Rdiv; ring.
specialize (jx_Betw xmin xmax x (fst dX) H1).
intros; auto with zarith.
intros; left; unfold convergence_error.
replace (jx (projT1 x - xmin) (fst dX)) with 0%Z.
rewrite zeroness_ubar_1 with (1:=usol_is_sol); auto with real.
rewrite zeroness_unh_1 with (2:=unh_is_discrete_sol dX); auto with real.
unfold Rdiv; ring.
specialize (jx_Betw xmin xmax x (fst dX) H1).
intros; auto with zarith.
apply OuP_is_EV_zero.
destruct LL1h_is_second_order with (2:=usol_is_sol) (3:=f_is_zero_at_beginning) (4:=M) (5:=H)
  as (C ,(alpha,(H2,(H3,H4)))).
auto with real.
exists C;exists alpha.
intuition.
unfold Xxt,Xjk,kt.
simpl.
replace (0 / snd dX) with (IZR 0).
rewrite Int_part_IZR.
simpl.
rewrite Rmult_0_l.
assert (Betw xmin xmax (xmin + IZR (jx (projT1 X - xmin) (fst dX)) * fst dX)).
unfold Betw; split.
apply Rle_trans with (xmin+IZR 0 * (fst dX))%R.
right; simpl; ring.
apply Rplus_le_compat_l; apply Rmult_le_compat_r; auto with real.
apply IZR_le; unfold jx; apply Int_part_pos.
unfold Rdiv; apply Rmult_le_pos; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify.
destruct X as (x,(H1',H2')); simpl; assumption.
apply Rplus_le_reg_l with (-xmin).
apply Rmult_le_reg_l with (/fst dX); auto with real.
destruct X as (x,(Y1,Y2)); simpl; unfold jx.
apply Rle_trans with (IZR (Int_part ((x - xmin) / fst dX))).
right; field; auto with real.
apply Rle_trans with ((x - xmin) / fst dX).
apply (proj1 (base_Int_part _)).
unfold Rdiv; rewrite Rmult_comm.
apply Rmult_le_compat_l; auto with real.
apply Rplus_le_reg_l with xmin; ring_simplify; assumption.
replace (xmin + IZR (jx (projT1 X - xmin) (fst dX)) * fst dX) with
   (projT1 (exist _ _ H8)) by reflexivity.
apply H4;auto.
unfold Rdiv;simpl;ring.
Qed. 

Lemma condinite1h_aux: forall (P:R*R-> Prop) h g,
  (forall dX, P dX -> 0 < fst dX) ->
  (forall dX, P dX -> ni_exact xmin xmax dX) ->
  OuP P (fun (x : {x : R | Betw xmin xmax x}) dX => Rsqr (h dX (jx (projT1 x-xmin) (fst dX)))) g -> 
  OuP P (fun (_:unit) dX => finite_squared_norm_dx (fst dX) (h dX)) g.
unfold OuP; intros P h g H HH (alp,(C,(H1,(H2,H3)))).
exists alp; exists (C*(xmax-xmin)).
split; trivial.
split.
apply Rmult_lt_0_compat; auto with real.
apply Rplus_lt_reg_r with (xmin); ring_simplify.
apply xminLtxmax.
intros.
apply Rle_trans with ((xmax-xmin)*(C*Rabs (g dX))).
rewrite Rabs_right.
apply finite_squared_norm_dx_majoration; auto with real.
apply Rmult_le_pos; auto with real.
apply Rabs_pos.
intros.
rewrite Rsqr_abs.
replace j with (jx ((xmin+IZR j*fst dX)-xmin) (fst dX)).
unfold Rsqr; rewrite <- Rabs_mult.
assert (Betw xmin xmax (((xmin + IZR j * fst dX)))).
split.
apply Rle_trans with (xmin+0);[auto with real|apply Rplus_le_compat_l].
apply Rmult_le_pos; auto with real.
replace 0%R with (IZR 0) by reflexivity.
apply IZR_le; apply H5.
apply Rle_trans with (xmin+IZR ((ni (fst dX))) * (fst dX))%R.
apply Rplus_le_compat_l; apply Rmult_le_compat_r; auto with real.
apply IZR_le.
destruct H5; unfold Zpred; omega.
rewrite HH; auto.
right; field.
specialize (H dX H4); auto with real.
replace (xmin + IZR j * fst dX - xmin) with (projT1 (exist _ _ H6) - xmin) by reflexivity.
apply H3; auto.
unfold jx.
replace ((xmin+IZR j * fst dX -xmin)/ fst dX) with (IZR j).
apply Int_part_IZR.
field.
apply Rgt_not_eq; apply H; trivial.
apply Rle_ge; apply finite_squared_norm_dx_pos.
left; apply H; trivial.
apply Rle_trans with ((xmax-xmin)*(C * Rabs (g dX)));[idtac|right; ring].
apply Rmult_le_compat_r; auto with real.
apply Rmult_le_pos; auto with real.
apply Rabs_pos.
Qed.


(* lemme 73-210709 *)
Lemma condinite1h: 
 is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3%nat usol ->
 is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4%nat usol ->
  Oupse
   (fun (_:unit) dX => finite_norm_dx (fst dX)
               (fun j => convergence_error xmin usol unh dX j 1 / (snd dX)))
   (fun dX => fst dX ^2 +snd dX^2).
intros.
unfold R_n.finite_norm_dx.
apply OuP_sqrt.
intros _ dX (T1,T2).
apply finite_squared_norm_dx_pos; auto with real.
apply condinite1h_aux.
intros dX (Y1,_); assumption.
intros dX (_,(_,Y1)); assumption.
apply OuP_Rsqr.
intros; apply Rle_0_sqr.
apply OuP_eq_trans_r with  (fun dX : R * R => (fst dX ^ 2 + snd dX ^ 2)).
intros.
rewrite sqrt_Rsqr; auto with real.
apply Rle_trans with (0+0); try apply Rplus_le_compat;auto with real.
unfold pow; simpl; rewrite Rmult_1_r; fold (Rsqr (fst dX)); auto with real.
unfold pow; simpl; rewrite Rmult_1_r; fold (Rsqr (snd dX)); auto with real.
apply condinite1j; auto.
Qed.

(*lemme 74-210709*)
Lemma energyiniteh:
 is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3%nat usol ->
 is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4%nat usol ->
 Oupse
  (fun (_:unit) dX => 
    sqrt (energy c (convergence_error xmin usol unh dX) dX 0%nat))
  (fun dX => fst dX ^2 +snd dX^2).
intros.
unfold one_d_wave_equation.energy.
apply OuP_eq_trans_l with
  (fun (_ : unit) (dX : R * R) =>
   sqrt
     (1 / 2 *
      finite_squared_norm_dx (fst dX)
        (scalar_multiply_n (1 / snd dX)
           ((fun j : Z => convergence_error xmin usol unh dX j 1))))).
intros _ dX (Y1,(Y2,Y3)).
rewrite finite_dot_product_dx_zero1ou2; auto with real.
apply f_equal.
ring_simplify.
apply f_equal.
apply finite_squared_norm_dx_eq; trivial.
intros.
unfold scalar_multiply_n, subtract_n.
simpl.
replace (convergence_error xmin usol unh dX i 0) with 0.
ring.
unfold convergence_error; simpl.
rewrite unh_0; auto with real.
ring.
auto with zarith.
intros i (J1,J2).
case (Zle_lt_or_eq _ _ J1); clear J1; intros J1.
left; unfold Ah.
unfold convergence_error; simpl.
repeat rewrite unh_0; auto with real zarith.
unfold Rdiv; ring.
right; rewrite <- J1.
unfold convergence_error; simpl.
rewrite zeroness_ubar_1 with (1:=usol_is_sol).
rewrite zeroness_unh_1 with (2:=unh_is_discrete_sol dX); auto with real.
apply OuP_eq_trans_l with
  (scalar_multiplyAA_fun _  _  (sqrt (/2))
    (fun (_ : unit) (dX : R * R) =>
            (finite_norm_dx  (fst dX)
               (fun j : Z => convergence_error xmin usol unh dX j 1 / snd dX)))).
intros.
unfold scalar_multiplyAA_fun, scalar_multiply_n.
unfold R_n.finite_norm_dx; rewrite <- sqrt_mult.
apply f_equal.
replace (/2) with (1/2) by field.
apply f_equal.
apply finite_squared_norm_dx_eq; trivial.
destruct H1 as (H00,H1); auto.
intros; unfold Rdiv; ring.
auto with real.
apply finite_squared_norm_dx_pos.
destruct H1 as (H00,H1); auto with real.
apply OuP_multiplyAA.
assert (0 < sqrt (/2)); auto with real.
apply sqrt_lt_R0; auto with real.
apply condinite1h; trivial.
Qed.


(** The sum of all truncation errors is small *)

(* consistance lemme 75-210709*)
Lemma sum_truncation_error:  
   is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3 usol  -> 
   is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4 usol  -> 
   Oupse
  (fun (t:{t|Under T t}) dX =>
    snd dX * sum_f 1 (kt (projT1 t) (snd dX)-1)
    (fun n => (finite_norm_dx (fst dX)
                   (fun j => truncation_error xmin xmax c usol u0h u1h fh dX j  (S n)))))
  (fun dX => (fst dX)^2+ (snd dX)^2).
intros.
apply OuP_sum_f_minus1_bis with
 (f:=fun dX n => 
        (finite_norm_dx (fst dX)
           (fun j : Z => truncation_error xmin xmax c usol u0h u1h fh dX j n))).
apply T_pos.
intros dX (T1,(T2,T3)); trivial.
assert (scheme_is_consistent_of_order xmin xmax T c usol u0h u1h fh 2%nat 2%nat).
apply consistency with partial_derive_firstvar partial_derive_secondvar; auto with real.
unfold scheme_is_consistent_of_order in H0.
apply OuP_sqrt; trivial.
intros.
apply finite_squared_norm_dx_pos.
left; apply H2.
Qed.


(** The energy of the convergence error is small *)

Lemma energie: (*81p44- 140909*)
   is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3 usol  -> 
   is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4 usol  -> 
 OupsCFLe 
  (fun (t:{t|Under T t}) dX => 
    sqrt (energy c (convergence_error xmin usol unh dX) dX
                            (kt (projT1 t) (snd dX) -1)%nat))
  (fun dX => fst dX ^2 +snd dX^2).
intros M J2.
apply OuP_maj with 
  (fun (t : {t : R | Under T t}) => add2_fun
     (fun (dX : R * R) =>
        sqrt (energy c (convergence_error xmin usol unh dX) dX  0%nat))
     (scalar_multiplyA_fun _ (sqrt 2 / (2*sqrt (2*eps-Rsqr eps)))
       (fun dX => (snd dX) 
             * sum_f 1  (kt (projT1 t) (snd dX) - 1)
                     (fun n => (finite_norm_dx (fst dX)
                               (fun j => truncation_error xmin xmax c usol u0h u1h fh dX j (S n))))))).
intros X dX (Y1,(Y2,(Y3,Y4))).
unfold add2_fun, scalar_multiplyA_fun.
repeat rewrite Rabs_right.
rewrite <- Rmult_assoc.
apply energy_majoration with (u0h := fun _ _ => 0)
  (u1h := (fun dX j0 => convergence_error xmin usol unh dX j0 1 / snd dX))
  (fh := (fun dX j n => truncation_error xmin xmax c usol u0h u1h fh dX j (S n))); auto with real.
unfold is_discrete_solution; intros.
apply convergence_error_is_discrete_sol; auto.
(* . *)
apply Rle_ge.
replace 0 with (0+0) by ring.
apply Rplus_le_compat.
apply sqrt_pos.
repeat apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat.
auto with real.
apply sqrt_lt_R0.
replace (2*eps - Rsqr eps) with (eps * (2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1.
apply eps_lt.
auto with real.
left; now apply Y2.
apply sum_f_pos.
intros; apply sqrt_pos.
apply Rle_ge; apply sqrt_pos.
apply OuP_is_EV_plus.
apply OuP_unit.
apply OuP_implies_OuP with 
  (P1:=(fun dX => 0 < fst dX /\ 0 < snd dX /\ ni_exact xmin xmax dX)).
clear; intuition.
apply energyiniteh; trivial.
apply OuP_multiplyAA.
assert (0 < 2); auto with real.
unfold Rdiv; apply Rmult_integral_contrapositive_currified.
assert (0 < sqrt 2); auto with real.
apply sqrt_lt_R0; auto with real.
apply Rinv_neq_0_compat.
apply Rmult_integral_contrapositive_currified; auto with real.
assert (0 < sqrt (2*eps-Rsqr eps)); auto with real.
apply sqrt_lt_R0; auto with real.
replace (2*eps - Rsqr eps) with (eps * (2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1.
apply eps_lt.
auto with real.
apply OuP_implies_OuP with (2:=sum_truncation_error M J2). 
intros dX (H1,(H2,(H3,H4))); split; try split; trivial.
Qed.


(** The scheme is convergent *)

Lemma convergence: 
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3%nat usol  -> 
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4%nat usol  -> 
 scheme_is_convergent_of_order xmin xmax T c usol unh eps 2%nat 2%nat.
intros.
unfold scheme_is_convergent_of_order.
apply OuP_maj with
  (fun (t : {t : R | Under T t}) (dX : R * R) =>
    (sqrt 2 / (sqrt (2*eps-Rsqr eps))) * (snd dX) *
    sum_f 1 (kt (projT1 t) (snd dX)) 
              (fun n => sqrt (energy c (convergence_error xmin usol unh dX) dX (n-1)%nat))).
intros X dX (H1,(H2,(H3,H4))).
rewrite Rabs_right.
2: apply Rle_ge; apply sqrt_pos.
case_eq (kt (projT1 X) (snd dX)).
intros G.
unfold sum_f; simpl.
unfold R_n.finite_norm_dx; rewrite finite_squared_norm_dx_zero; auto.
rewrite sqrt_0.
apply Rabs_pos.
intros; unfold convergence_error.
rewrite unh_0; auto with real zarith.
intros n G.
apply Rle_trans with 
 (sum_f 1 (S n)
     (fun n : nat =>  (finite_norm_dx (fst dX)
         (fun j => convergence_error xmin usol unh dX j n
                           - convergence_error xmin usol unh dX j (n-1))))).
apply Rle_trans with
  ((finite_norm_dx (fst dX)
     (fun j => (sum_f 1 (S n)
     (fun n : nat  => convergence_error xmin usol unh dX j n
                           - convergence_error xmin usol unh dX j (n-1)))))).
right.
unfold R_n.finite_norm_dx; apply f_equal.
apply finite_squared_norm_dx_eq; auto with real.
intros.
clear G; induction n.
unfold sum_f; simpl.
unfold convergence_error at 3.
rewrite unh_0; auto with real zarith.
ring.
unfold sum_f; simpl.
unfold sum_f in IHn; simpl in IHn.
replace (n-0)%nat with n in IHn by auto with arith.
rewrite <- IHn.
replace (n+1-0)%nat with (n+1)%nat by auto with arith.
replace (S n) with (n+1)%nat.
ring.
omega.
clear G; induction n.
unfold sum_f; simpl.
apply Rle_refl.
unfold sum_f in *; simpl in *.
eapply Rle_trans.
apply finite_squared_norm_dx_triang; auto.
auto with real.
apply Rplus_le_compat_r.
replace (n-0)%nat with n in IHn by auto with arith.
apply IHn.
rewrite Rabs_right.
clear G; induction n.
unfold sum_f; simpl.
change 1%nat with (0+1)%nat.
apply minorationenergy2 with (fun _ _ => 0) 
   (fun dX j0 => convergence_error xmin usol unh dX j0 1 / snd dX)
    (fun dX (j0 : Z) (m : nat) =>
          truncation_error xmin xmax c usol u0h u1h fh dX j0 (S m)); auto with real.
unfold is_discrete_solution; intros.
apply convergence_error_is_discrete_sol; auto with real.
unfold sum_f in *; simpl in *.
rewrite Rmult_plus_distr_l.
apply Rplus_le_compat.
replace (n-0)%nat with n in IHn by auto with arith.
apply IHn.
replace (S (n + 1)) with ((n+1-0)+1)%nat.
apply minorationenergy2 with (fun _ _ => 0) 
   (fun dX j0 => convergence_error xmin usol unh dX j0 1 / snd dX)
    (fun dX (j0 : Z) (m : nat) =>
          truncation_error xmin xmax c usol u0h u1h fh dX j0 (S m)); auto with real.
unfold is_discrete_solution; intros.
apply convergence_error_is_discrete_sol; auto with real.
omega.
apply Rle_ge; apply Rmult_le_pos.
apply Rmult_le_pos.
unfold Rdiv; apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
apply sqrt_lt_R0.
replace (2*eps-Rsqr eps) with (eps*(2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps.
ring_simplify.
apply Rlt_trans with 1; try apply eps_lt; auto with real.
now left.
unfold sum_f; simpl.
apply cond_pos_sum.
intros; apply sqrt_pos.
apply OuP_eq_trans_l with (scalar_multiplyAA_fun _ _
  (sqrt 2 / sqrt (2*eps-Rsqr eps))
  (fun (t : {t : R | Under T t}) (dX : R * R) =>
    snd dX *
    sum_f 1 (kt (projT1 t) (snd dX))
     (fun n : nat =>
      sqrt (energy c (convergence_error xmin usol unh dX) dX (n - 1)%nat)))).
intros.
unfold scalar_multiplyAA_fun; ring.
apply OuP_multiplyAA.
assert (0 < 2); auto with real.
unfold Rdiv; apply Rmult_integral_contrapositive_currified.
assert (0 < sqrt 2); auto with real.
apply sqrt_lt_R0; auto with real.
apply Rinv_neq_0_compat.
assert (0 < sqrt (2*eps-Rsqr eps)); auto with real.
apply sqrt_lt_R0; auto with real.
replace (2*eps - Rsqr eps) with (eps * (2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1; auto with real.
apply eps_lt.
apply OuP_sum_f.
apply T_pos.
intros dH (_,(K,_)); exact K.
apply energie; trivial.
Qed.

Lemma convergence_xt: 
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 3%nat usol  -> 
  is_sufficiently_regular_n xmin xmax partial_derive_firstvar partial_derive_secondvar 4%nat usol  ->  
  OupsCFLe 
  (fun (X: {x : R | Betw xmin xmax x} * {t : R | Under T t}) (dX : R * R) 
        => convergence_error xmin usol unh dX (jx (projT1 (fst X)-xmin) (fst dX)) (kt (projT1 (snd X)) (snd dX)))
  (fun dX : R * R => (fst dX ^ 2 + snd dX ^ 2) / sqrt (fst dX)).
intros.
specialize (convergence H H0); unfold scheme_is_convergent_of_order.
intros (alp,(C, (Y1,(Y2,Y3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX H1 (H2,(H3,(H4,H5))).
unfold Rdiv; rewrite Rabs_mult, <- Rmult_assoc.
rewrite (Rabs_right (/ sqrt (fst dX))).
2: apply Rle_ge; left; apply Rinv_0_lt_compat.
2: apply sqrt_lt_R0; auto.
specialize (jx_Betw _ _ (fst X) _ H2).
intros H6.
case (Zle_lt_or_eq (jx (projT1 (fst X)-xmin) (fst dX)) (ni (fst dX))); auto with zarith.
intros.
apply finite_norm_le_reverse with (xmin:=xmin) (xmax:=xmax) 
  (uh:=(fun j => convergence_error xmin usol unh dX j (kt (projT1 (snd X)) (snd dX)))); auto with zarith.
rewrite <- (Rabs_right (R_n.finite_norm_dx xmin xmax (fst dX)
  (fun j0 : Z =>
   convergence_error xmin usol unh dX j0 (kt (projT1 (snd X)) (snd dX))))).
apply Y3; auto.
apply Rle_ge; unfold R_n.finite_norm_dx.
apply Reals_compl.sqrt_pos.
intros Y; rewrite Y.
replace (convergence_error xmin usol unh dX (ni (fst dX))
     (kt (projT1 (snd X)) (snd dX))) with 0%R.
rewrite Rabs_R0; apply Rmult_le_pos;try apply Rmult_le_pos; auto with real.
apply Rabs_pos.
left; apply Rinv_0_lt_compat; apply sqrt_lt_R0; assumption.
unfold convergence_error.
rewrite zeroness_unh_2 with (2:=unh_is_discrete_sol dX); auto with real.
rewrite zeroness_ubar_2 with (1:=usol_is_sol); auto with real.
Qed.

End Conv.
Extraction Inline OuP_sum_f energie energyiniteh sum_truncation_error
 condinite1h OuP_sum_f_minus1_bis consistency condinite1j 
  Oup_pointwise_to_squared_norm_dx_truncation_error 
  pointwise_consistency DL_discretization_eps_dt 
  OuP_case Oup_pointwise_to_squared_norm_dx OuP_match
  LL1h_is_second_order DL_discretization_eps_aux
  DL_discretization_eps_dx OuP_dX_to_eps_Betw.

Recursive Extraction convergence.

