(** Stability of the numerical scheme *)

Require Import Fourier.
Require Export one_d_wave_equation.

Open Local Scope nat_scope.
Open Local Scope R_scope.

Section Stab.

(** Here, we take one particular dX for the whole file.
    It will go towards zero only in convergence.v. *)

Variable xmin xmax:R.
Hypothesis xminLtxmax: xmin < xmax.

Variable c:R.
Variable u0h u1h: R*R -> Z-> R.
Variable fh: R*R -> Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable usol : R * R -> R.
Variable eps:R.
Variable dX:R*R.

Hypothesis unh_is_discrete_sol: is_discrete_solution xmin xmax c u0h u1h fh unh dX.
Hypothesis eps_lt: 0 < eps < 1.
Hypothesis CFL_eps: CFL_condition_r c eps dX.
Hypothesis c_pos: 0 < c.


Notation ni x := (ni xmin xmax x).
Notation finite_dot_product_dx := (fun dx u v => (finite_dot_product_dx xmin xmax dx u v)).
Notation finite_squared_norm_dx := (fun dx u => (finite_squared_norm_dx xmin xmax dx u)).
Notation finite_norm_dx := (fun dx u => (finite_norm_dx xmin xmax dx u)).
Notation energy := (fun c uh dX n => energy xmin xmax c uh dX n).

(** Evolution of the energy when time increases *)

Lemma energy_increases: forall  n:nat,
   0 < fst dX -> snd dX <> 0 -> (0<n)%nat->
   (energy c (unh dX) dX n) - (energy c (unh dX) dX (n-1)%nat) =
  1/2*finite_dot_product_dx (fst dX) (fun j => fh dX j n) 
            (fun j => unh dX j (n+1)%nat - unh dX j (n-1)%nat).
intros n HH H1 H2.
rewrite finite_dot_product_dx_eq_r with (xmin:=xmin) (xmax:=xmax)
 (v:=subtract_n (fun j => unh dX j (n+1)%nat)  (fun j => unh dX j (n-1)%nat)); auto.
unfold one_d_wave_equation.energy.
repeat rewrite <- Rmult_plus_distr_l in |- *.      
rewrite <- Rmult_minus_distr_l in |- *. 
apply Rmult_eq_compat_l.
unfold R_n.finite_squared_norm_dx.
rewrite finite_dot_product_dx_eq_strict_l with (u:=(fun j : Z => fh dX j n)) 
  (v:=(add_n (scalar_multiply_n (1/(snd dX*snd dX)) 
             (subtract_n (add_n (fun j =>unh dX j (n+1)%nat) 
                            (fun j =>unh dX j (n-1)%nat))
                       (scalar_multiply_n 2 (fun j=> unh dX j n)))) 
               (fun j => (Ah c (fun j' => unh dX j' n) (fst dX) j)))); auto.
repeat rewrite  finite_dot_product_dx_add_l.
repeat rewrite finite_dot_product_dx_scal_mult_l.
repeat rewrite finite_dot_product_dx_scal_mult_r.
repeat rewrite  finite_dot_product_dx_sub_l.
repeat rewrite finite_dot_product_dx_sub_r.
repeat rewrite  finite_dot_product_dx_add_l.
replace (n - 1 +1)%nat with n;try omega.
repeat rewrite finite_dot_product_dx_scal_mult_l.
repeat rewrite  finite_dot_product_dx_sub_l.
repeat rewrite  finite_dot_product_dx_sub_r.
field_simplify;try assumption.
repeat rewrite (finite_dot_product_dx_sym _ _ (fst dX) (fun j : Z => unh dX j n)).
repeat rewrite (finite_dot_product_dx_sym _ _ (fst dX) _ (fun j : Z => unh dX j (n-1))).
rewrite (finite_dot_product_dx_sym _ _ (fst dX) (fun j => unh dX j (n+1))  (fun j : Z => unh dX j (n-1))).
field_simplify;try assumption.
unfold Rdiv;apply Rmult_eq_compat_r.
unfold Rminus;repeat rewrite Rplus_assoc.
repeat apply Rplus_eq_compat_l.
apply Rplus_eq_compat_r.
repeat apply f_equal.
rewrite finite_dot_product_dx_eq_l with (v:= Ah c (fun j' : Z => unh dX j' (n - 1)) (fst dX))
(u:=fun j : Z => Ah c (fun j' : Z => unh dX j' (n - 1)) (fst dX) j); auto.
rewrite Ah_is_symmetrical with _ _ c (fst dX) (fun j'=> unh dX j' (n-1))
  (fun j : Z => unh dX j n); auto.
apply finite_dot_product_dx_sym.
rewrite zeroness_unh_1 with (2:=unh_is_discrete_sol); auto with zarith.
rewrite zeroness_unh_2 with (2:=unh_is_discrete_sol); auto with zarith.
rewrite zeroness_unh_1 with (2:=unh_is_discrete_sol); auto with zarith.
rewrite zeroness_unh_2 with (2:=unh_is_discrete_sol); auto with zarith.
intros i Hi.
unfold add_n,scalar_multiply_n,subtract_n;simpl.
assert (exists m, n = S m).
exists (pred n); omega. 
destruct H as (m, Hm).	
repeat rewrite Hm.
replace (S m +1)%nat with (m+2)%nat by omega.
replace (S m -1)%nat with m by omega.
repeat replace (S m) with (m+1)%nat by omega.
intros.
generalize (unh_is_discrete_sol H1 i (S (S m))); simpl.
intros T.
rewrite <- Rplus_0_l.
rewrite <- Ropp_0.
replace (S m) with (m+1)%nat in T by omega.
rewrite <- T.
apply Rplus_eq_reg_l with (-fh dX i (m+1)); ring_simplify.
rewrite scheme_eq_3; auto.
pattern (m+1)%nat at 1 2; replace (m+1)%nat with (S m)%nat by omega.
unfold Lh; field; assumption.
split; auto with zarith.
unfold subtract_n.
repeat rewrite zeroness_unh_1 with (2:=unh_is_discrete_sol); try assumption; auto with real zarith.
Save.

(** Energy is greater than... *)
Lemma minorationenergy:  forall n,
   snd dX <> 0 -> 0 < (fst dX)  ->
    1/2*(1-Rsqr (CFL c dX))*
  finite_squared_norm_dx (fst dX) 
       (fun j=> 1/(snd dX)*(unh dX j (n+1)%nat - unh dX j n))
  <= energy c (unh dX) dX n.
intros.
unfold one_d_wave_equation.energy.
rewrite <- Rmult_plus_distr_l.
rewrite Rmult_assoc.
apply Rmult_le_compat_l.
fourier.
unfold Rminus at 1; rewrite Rmult_plus_distr_r.
apply Rplus_le_compat.
rewrite Rmult_1_l.
apply Rle_refl.
rewrite semi_finite_dot_product_dx with _ _ (fst dX)
   (fun u j => Ah c u (fst dX) j)  (fun j' : Z => unh dX j' n)
   (fun j : Z => unh dX j (n + 1)); auto.
2: apply Ah_is_symmetrical; auto.
2: apply  zeroness_unh_1 with (2:=unh_is_discrete_sol); auto with zarith.
2: apply  zeroness_unh_2 with (2:=unh_is_discrete_sol); auto with zarith.
2: apply  zeroness_unh_1 with (2:=unh_is_discrete_sol); auto with zarith.
2: apply  zeroness_unh_2 with (2:=unh_is_discrete_sol); auto with zarith.
unfold subtract_n, add_n.
apply Rle_trans with (-( /4 *
 finite_dot_product_dx (fst dX)  
  (fun j : Z => Ah c (fun i : Z => unh dX i n - unh dX i (n + 1)) (fst dX) j)
   (fun i : Z => unh dX i n - unh dX i (n + 1) ))).
rewrite Ropp_mult_distr_l_reverse.
apply Ropp_le_contravar.
apply Rmult_le_reg_l with 4.
fourier.
rewrite <- Rmult_assoc.
rewrite Rinv_r.
2: discrR.
rewrite Rmult_1_l.
apply Rle_trans with
  (4 * c * c / ((fst dX) * (fst dX)) * finite_squared_norm_dx (fst dX)
     (fun i : Z => unh dX i n - unh dX i (n + 1))).
apply Ah_cont; auto.
repeat rewrite zeroness_unh_1 with (2:=unh_is_discrete_sol); auto with zarith real.
repeat rewrite zeroness_unh_2 with (2:=unh_is_discrete_sol); auto with zarith real.
rewrite finite_squared_norm_dx_eq with 
  (f2:=scalar_multiply_n  (-1 / snd dX)
      (fun i : Z => unh dX i n - unh dX i (n + 1)))
  (f1:= (fun j : Z => 1 / snd dX * 
     (unh dX j (n + 1) - unh dX j n))); auto.
rewrite finite_squared_norm_dx_homogen.
unfold CFL.
right; unfold Rsqr; field.
split; trivial; now apply Rgt_not_eq.
intros; unfold scalar_multiply_n.
field.
exact H.
rewrite  <- (Rplus_0_l  (-(/ 4 *
  finite_dot_product_dx (fst dX) 
     (fun j : Z => Ah c (fun i : Z => unh dX i n - unh dX i (n + 1)) (fst dX) j)
     (fun i : Z => unh dX i n - unh dX i (n + 1) )))).
apply Rplus_le_compat_r.
apply Rmult_le_pos.
fourier.
apply Ah_pos; auto.
repeat rewrite zeroness_unh_1 with (2:=unh_is_discrete_sol); auto with zarith real.
repeat rewrite zeroness_unh_2 with (2:=unh_is_discrete_sol); auto with zarith real.
Qed.



Lemma energy_pos: 
 forall n, (0 < snd dX) -> 0 < (fst dX)  ->
   0 <= energy c (unh dX) dX n.
intros.
assert (snd dX <> 0) by auto with real.
apply Rle_trans with ( 1/2*(1-Rsqr (CFL c dX))*
  finite_squared_norm_dx (fst dX) 
       (fun j=> 1/(snd dX)*(unh dX j (n+1)%nat - unh dX j n))).
2: apply minorationenergy; trivial.
apply Rmult_le_pos.
repeat apply Rmult_le_pos.
auto with real.
fourier.
apply Rplus_le_reg_l with (Rsqr (CFL c dX)); ring_simplify.
replace 1 with (Rsqr 1) by (unfold Rsqr; ring).
apply Rsqr_incr_1.
apply Rle_trans with (1-eps).
apply CFL_eps.
apply Rplus_le_reg_l with (-1+eps); ring_simplify.
left; apply eps_lt.
unfold CFL; repeat apply Rmult_le_pos; trivial.
now left.
now left.
left; now apply Rinv_0_lt_compat.
auto with real.
apply finite_squared_norm_dx_pos.
now left.
Qed.


(* old version 
Lemma minorationenergy2: (*73p38-140909*)
  forall n,
   sqrt
     (squared_norm_dx xmin xmax (fst dX)
        (fun j : Z =>
          unh dX j (n+1) -
          unh dX j n)) <=     (* was convergence_error *)
   (sqrt 2 / sqrt (1 - Rsqr (CFL c dX)) * snd dX *
       sqrt (energy c (unh dX) dX xmin xmax n)). *)

Lemma minorationenergy2: (*73p38-140909*)
  forall n,
  0 < fst dX -> 0 < snd dX ->
     finite_norm_dx (fst dX)
        (fun j : Z =>
          unh dX j (n+1) -
          unh dX j n) <=     
   (sqrt 2 / sqrt (2*eps - Rsqr eps) * snd dX *
       sqrt (energy c (unh dX) dX n)).
intros.
assert (snd dX <> 0) by auto with real.
assert (0 < 2*eps - Rsqr eps).
replace (2*eps - Rsqr eps) with (eps*(2-eps)) by (unfold Rsqr;ring). 
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1.
apply eps_lt.
auto with real.
apply Rsqr_incr_0_var.
unfold R_n.finite_norm_dx; rewrite Rsqr_sqrt.
repeat rewrite Rsqr_mult.
rewrite Rsqr_sqrt.
apply Rmult_le_reg_l with ((1 / 2 * (1 - Rsqr (CFL c dX)))/(Rsqr (snd dX))). 
unfold Rdiv; repeat apply Rmult_lt_0_compat.
auto with real.
auto with real.
apply Rplus_lt_reg_r with (Rsqr (CFL c dX)); ring_simplify.
rewrite <- Rsqr_1.
apply Rsqr_incrst_1.
apply Rle_lt_trans with (1-eps).
apply CFL_eps.
apply Rplus_lt_reg_r with (-1+eps); ring_simplify.
apply eps_lt.
unfold CFL.
unfold Rdiv; repeat apply Rmult_le_pos; auto with real.
auto with real.
auto with real.
apply Rle_trans with (1 / 2 * (1 - (CFL c dX)²) *
       finite_squared_norm_dx (fst dX)
         (fun j : Z => 1 / snd dX * (unh dX j (n + 1) - unh dX j n))).
right.
unfold Rdiv; rewrite Rmult_assoc.
apply f_equal.
replace (/ (Rsqr (snd dX))) with (/(snd dX) * / (snd dX)).
rewrite <- finite_squared_norm_dx_homogen.
apply finite_squared_norm_dx_eq; auto.
intros; unfold scalar_multiply_n.
ring.
unfold Rsqr; field; auto with real.
apply Rle_trans with (energy c (unh dX) dX n).
apply minorationenergy; trivial.
rewrite <- (Rmult_1_l (energy c (unh dX) dX n)) at 1.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
apply energy_pos; trivial.
rewrite Rsqr_div.
repeat rewrite Rsqr_sqrt.
apply Rle_trans with ((1 - (CFL c dX)²) / (2 * eps - eps²)).
2: right; field.
apply Rmult_le_reg_l with (2 * eps - eps²).
exact H2.
rewrite Rmult_1_r.
apply Rle_trans with ((1 - (CFL c dX)²)).
unfold CFL_condition_r in CFL_eps.
apply Rle_trans with (1-Rsqr (1-eps)).
unfold Rsqr; right; ring.
unfold Rminus; apply Rplus_le_compat_l.
apply Ropp_le_contravar.
apply Rsqr_incr_1.
exact CFL_eps.
unfold CFL.
unfold Rdiv; repeat apply Rmult_le_pos; auto with real.
apply Rplus_le_reg_l with eps; ring_simplify.
left; apply eps_lt.
right; field.
now apply Rgt_not_eq.
split.
now apply Rgt_not_eq.
apply Rgt_not_eq.
apply Rlt_gt; auto with real.
now left.
auto with real.
apply Rgt_not_eq.
now apply sqrt_lt_R0.
apply energy_pos.
exact H0.
exact H.
apply finite_squared_norm_dx_pos.
now left.
repeat apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
now apply sqrt_lt_R0.
auto with real.
apply sqrt_pos.
Qed.

(** Energy is smaller than... *)
Lemma energy_majoration_aux: 
 forall n, (0 < n)%nat -> 0 < snd dX-> 0 < fst dX ->
  sqrt (energy c (unh dX) dX n)
   <= sqrt (energy c (unh dX) dX (n-1)%nat)
     +  (sqrt 2 / (2*sqrt (2*eps-Rsqr eps)))
           * (snd dX) * (finite_norm_dx (fst dX) 
                               (fun j => fh dX j n)).
intros n Hn HdX2 HdX1.
assert (K:snd dX <> 0) by auto with real.
apply Rplus_le_reg_l with 
  (-sqrt (energy c (unh dX) dX (n - 1)%nat)); ring_simplify.
assert (0 <= sqrt (energy c (unh dX) dX n)).
apply sqrt_pos.
case H; intros H1; clear H.
apply Rmult_le_reg_l with
  (sqrt (energy c (unh dX) dX (n - 1)%nat) +
     sqrt (energy c (unh dX) dX n)).
rewrite <- (Rplus_0_l 0).
apply Rle_lt_trans with (sqrt (energy c (unh dX) dX (n - 1)%nat) + 0).
apply Rplus_le_compat_r.
apply sqrt_pos.
now apply Rplus_lt_compat_l.
apply Rle_trans with (energy c (unh dX) dX n
 - energy c (unh dX) dX (n - 1)%nat).
right; apply trans_eq with
   (Rsqr(sqrt (energy c (unh dX) dX n))
  -Rsqr(sqrt (energy c (unh dX) dX (n - 1)%nat))).
unfold Rsqr; ring.
repeat rewrite Rsqr_sqrt; try reflexivity.
apply energy_pos; trivial.
apply energy_pos; trivial.
rewrite energy_increases; trivial.
apply Rle_trans with
  (1 / 2 *
 ( (finite_norm_dx (fst dX) 
       (fun j => unh dX j (n + 1) - unh dX j (n - 1)))
  * (finite_norm_dx (fst dX) 
  (fun j : Z => fh dX j n)))).
apply Rmult_le_compat_l.
fourier.
apply Rle_trans with (1:=RRle_abs _ ).
apply Rle_trans with (finite_norm_dx (fst dX) (fun j : Z => fh dX j n)
   * finite_norm_dx (fst dX) (fun j : Z => unh dX j (n + 1) - unh dX j (n - 1))).
apply finite_Cauchy_Schwarz_dx.
now left.
right; ring.
repeat rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
apply sqrt_pos.
apply Rmult_le_reg_l with 2.
auto with real.
rewrite <- Rmult_assoc.
replace (2*(1/2)) with 1 by field.
rewrite Rmult_1_l.
apply Rle_trans with
 ((finite_norm_dx  (fst dX)
     (fun j : Z => unh dX j n - unh dX j (n-1)))
  +  ((finite_norm_dx (fst dX)
     (fun j : Z => unh dX j (n + 1) - unh dX j n)))).
apply Rle_trans with 
 (finite_norm_dx (fst dX) (fun j : Z => (unh dX j n - unh dX j (n - 1)) +
                 (unh dX j (n + 1) - unh dX j n))).
2: apply finite_squared_norm_dx_triang.
unfold R_n.finite_norm_dx; right; apply f_equal.
apply finite_squared_norm_dx_eq; auto.
intros; ring.
now left.
apply Rle_trans with
  (sqrt 2 / sqrt (2*eps - Rsqr eps) * snd dX *
        sqrt (energy c (unh dX) dX (n-1)%nat)
 + sqrt 2 / sqrt (2*eps - Rsqr eps) * snd dX *
        sqrt (energy c (unh dX) dX n)).
apply Rplus_le_compat.
pattern n at 1; replace n with ((n-1)+1)%nat.
apply minorationenergy2; trivial.
omega.
apply minorationenergy2; trivial.
right; field.
apply Rgt_not_eq.
apply sqrt_lt_R0.
replace (2*eps-Rsqr eps) with (eps *(2-eps)) by (unfold Rsqr;ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1; try apply eps_lt; auto with real.
rewrite <- H1.
apply Rle_trans with 0.
rewrite Rplus_0_r.
rewrite <- Ropp_0.
apply Ropp_le_contravar.
apply sqrt_pos.
unfold Rdiv; repeat apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat.
auto with real.
apply sqrt_lt_R0.
replace (2*eps-Rsqr eps) with (eps *(2-eps)) by (unfold Rsqr;ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1; try apply eps_lt; auto with real.
now left.
apply sqrt_pos.
Qed.



Lemma energy_majoration: 
 0 < snd dX-> 0 < fst dX ->
  forall t,
    let k:= (kt t (snd dX) -1)%nat in 
    sqrt (energy c (unh dX) dX k) 
       <= sqrt (energy c (unh dX) dX 0%nat)
               +  (sqrt 2 / (2*sqrt (2*eps-Rsqr eps)))  * (snd dX) * sum_f 1 k 
                     (fun n => (finite_norm_dx (fst dX) 
                               (fun j => fh dX j n))). 
intros.
induction k.
unfold sum_f; simpl.
ring_simplify.
apply Rle_refl.
apply Rle_trans with (1:= energy_majoration_aux (S k) (lt_O_Sn _)  H H0).
replace (S k -1)%nat with k by omega.
rewrite sum_f_S.
2: omega.
rewrite Rmult_plus_distr_l.
rewrite <- Rplus_assoc.
apply Rplus_le_compat_r.
exact IHk.
Qed.


Lemma energy_majoration_zero: 
 0 < snd dX-> 0 < fst dX ->
  (forall j k, fh dX j k =0) ->
  forall t,
    let k:= (kt t (snd dX) -1)%nat in 
    sqrt (energy c (unh dX) dX k) 
       <= sqrt (energy c (unh dX) dX 0%nat).
intros.
apply Rle_trans with (1:=energy_majoration H H0 t).
rewrite <- Rplus_0_r.
apply Rplus_le_compat_l.
right.
replace (sum_f 1 (kt t (snd dX) - 1)
  (fun n : nat => finite_norm_dx (fst dX) (fun j : Z => fh dX j n))) with 0%R.
ring.
induction (kt t (snd dX) - 1)%nat.
unfold sum_f; now simpl.
rewrite sum_f_S; auto with arith.
rewrite <- IHn.
unfold R_n.finite_norm_dx.
rewrite finite_squared_norm_dx_zero; auto.
rewrite sqrt_0; ring.
Qed.

End Stab.


(*
Section StabFP.
(** Useless for convergence, useful to prove each scheme result is bounded
      to gurantee the rounding errors are bounded *)
Variable xmin xmax:R.
Hypothesis xminLtxmax: xmin < xmax.

Variable c:R.
Variable u0: R->R.
Let u0h := (fun dX j => u0 (fst (Xjk xmin dX j 0))).

Variable u1h: R*R -> Z-> R.
Variable fh: R*R -> Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable usol : R * R -> R.
Variable eps:R.
Variable dX:R*R.

Hypothesis unh_is_discrete_sol: is_discrete_solution xmin xmax c u0h u1h fh unh dX.
Hypothesis eps_lt: 0 < eps < 1.
Hypothesis CFL_eps: CFL_condition_r c eps dX.
Hypothesis c_pos: 0 < c.


Notation ni x := (ni xmin xmax x).
Notation finite_dot_product_dx := (fun dx u v => (finite_dot_product_dx xmin xmax dx u v)).
Notation finite_squared_norm_dx := (fun dx u => (finite_squared_norm_dx xmin xmax dx u)).
Notation finite_norm_dx := (fun dx u => (finite_norm_dx xmin xmax dx u)).
Notation energy := (fun c uh dX n => energy xmin xmax c uh dX n).





Lemma unh_bounded: forall n, (Z_of_nat n <= ni (fst dX))%Z -> 0 < fst dX -> 0 < snd dX ->
     u0h dX 0 = 0 -> (forall j k, fh dX j k =0) ->
    finite_norm_dx (fst dX) (fun j => unh dX j n) <= finite_norm_dx (fst dX) (u0h dX) + 
          ((xmax-xmin)/c * (1-eps)) * (sqrt 2 / sqrt (2 * eps - Rsqr eps)) * sqrt (energy c (unh dX) dX 0%nat).
intros n Hn H1 H2 H0 Hf; destruct n.
(* n = 0 *)
rewrite <- (Rplus_0_r (finite_norm_dx (fst dX) (fun j => unh dX j 0))).
apply Rplus_le_compat.
right; unfold R_n.finite_norm_dx; apply f_equal.
apply finite_squared_norm_dx_eq; auto.
intros i Hi; generalize (unh_is_discrete_sol i 0%nat).
unfold scheme, Lh.
case (Z_lt_le_dec 0 i).
case (Z_lt_le_dec i (ni (fst dX))).
intros.
replace (unh dX i 0) with ((unh dX i 0 - u0h dX i)+u0h dX i) by ring.
rewrite H; auto with zarith.
ring.
intros J1 _ _; contradict J1; auto with zarith.
intros J1; replace i with 0%Z; auto with zarith.
intros Y; rewrite Y.
now rewrite H0.
split; auto with zarith.
unfold Rdiv; repeat apply Rmult_le_pos.
apply Rplus_le_reg_l with xmin; ring_simplify.
left; apply xminLtxmax.
left; auto with real.
apply Rplus_le_reg_l with eps; ring_simplify; left; apply eps_lt.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
apply sqrt_lt_R0; auto.
replace (2*eps - Rsqr eps) with (eps*(2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_le_trans with 1;[apply eps_lt|auto with real].
apply sqrt_pos.
(* n > 0  *)
apply Rle_trans with 
  (finite_norm_dx (fst dX) (u0h dX) + sum_f 0 n
     (fun n : nat =>  (finite_norm_dx (fst dX)
         (fun j => unh dX j (n+1) - unh dX j n)))).
apply Rle_trans with
  ((finite_norm_dx (fst dX)
     (fun j => (u0h dX j + sum_f 0 n
     (fun n : nat  => unh dX j (n+1)- unh dX j n))))).
(* + *)
right.
unfold R_n.finite_norm_dx; apply f_equal.
apply finite_squared_norm_dx_eq; auto with real.
intros; induction n.
unfold sum_f; simpl.
generalize (unh_is_discrete_sol i 0%nat).
unfold scheme, Lh.
case (Z_lt_le_dec 0 i).
case (Z_lt_le_dec i (ni (fst dX))).
intros.
replace (unh dX i 0) with ((unh dX i 0 - u0h dX i)+u0h dX i) by ring.
rewrite H3; auto with zarith.
ring.
intros J1 _ _; contradict J1; auto with zarith.
intros J1; replace i with 0%Z; auto with zarith.
intros Y; rewrite Y.
rewrite H0; ring.
auto with zarith.
unfold sum_f; simpl.
assert ((Z_of_nat (S n) <= ni (fst dX))%Z).
apply Zle_trans with (2:=Hn).
repeat rewrite inj_S; auto with zarith arith.
specialize (IHn H3).
unfold sum_f in IHn; simpl in IHn.
replace (n-0)%nat with n in IHn by auto with arith.
rewrite <- Rplus_assoc.
rewrite <- IHn.
replace (n+0)%nat with (n)%nat by auto with arith.
replace (S n) with (n+1)%nat.
ring.
omega.
(* + *)
apply Rle_trans with (finite_norm_dx (fst dX) (u0h dX) +
   finite_norm_dx (fst dX) (fun j => sum_f 0 n (fun n0 : nat => unh dX j (n0 + 1) - unh dX j n0))).
apply finite_squared_norm_dx_triang; auto with real.
apply Rplus_le_compat_l.
induction n.
unfold sum_f; simpl.
apply Rle_refl.
assert ((Z_of_nat (S n) <= ni (fst dX))%Z).
apply Zle_trans with (2:=Hn).
repeat rewrite inj_S; auto with zarith arith.
specialize (IHn H).
unfold sum_f in *; simpl in *.
eapply Rle_trans.
apply finite_squared_norm_dx_triang; auto.
auto with real.
apply Rplus_le_compat_r.
replace (n-0)%nat with n in IHn by auto with arith.
apply IHn.
(* + *)
apply Rplus_le_compat_l.
apply Rle_trans with (sum_f 0 n
  (fun n0 : nat => sqrt 2 / sqrt (2 * eps - eps²) * snd dX *
       sqrt (energy c (unh dX) dX n0))).
clear Hn; induction n.
unfold sum_f; simpl.
apply Rle_trans with (R_n.finite_norm_dx xmin xmax (fst dX)
         (fun j : Z => unh dX j (0 + 1)%nat - unh dX j 0)).
2: apply minorationenergy2 with u0h u1h fh; auto.
right; unfold R_n.finite_norm_dx; apply f_equal.
apply finite_squared_norm_dx_eq; auto.
unfold sum_f in *; simpl in *.
apply Rplus_le_compat.
replace (n-0)%nat with n in IHn by omega.
exact IHn.
replace (n+0)%nat with n by omega.
apply Rle_trans with (R_n.finite_norm_dx xmin xmax (fst dX)
         (fun j : Z => unh dX j (S n + 1)%nat - unh dX j (S n))).
2: apply minorationenergy2 with u0h u1h fh; auto.
right; unfold R_n.finite_norm_dx; apply f_equal.
apply finite_squared_norm_dx_eq; auto.
(* ++ *)
assert (0 < 2*eps - Rsqr eps).
replace (2*eps - Rsqr eps) with (eps*(2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_le_trans with 1;[apply eps_lt|auto with real].
apply Rle_trans with ((INR (n - 0) + 1) * 
  (sqrt 2 / sqrt (2 * eps - eps²) * snd dX *sqrt (energy c (unh dX) dX 0%nat)))%R.
apply Rle_trans with (sum_f 0 n
  (fun n0 : nat =>
   Rabs (sqrt 2 / sqrt (2 * eps - eps²) * snd dX * sqrt (energy c (unh dX) dX n0)))).
clear Hn; induction n.
unfold sum_f; simpl.
apply RRle_abs.
unfold sum_f in *; simpl in *.
apply Rplus_le_compat.
replace (n-0)%nat with n in IHn by omega.
exact IHn.
apply RRle_abs.
apply sum_f_le.
intros.
rewrite Rabs_right.
apply Rmult_le_compat_l.
unfold Rdiv; repeat apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
now apply sqrt_lt_R0.
now left.
replace i with (kt (INR (i+1)*snd dX) (snd dX) - 1)%nat.
apply energy_majoration_zero with u0h u1h fh eps; auto.
unfold kt.
replace (INR (i+1) * snd dX / snd dX) with (INR (i+1)).
rewrite Int_part_INR, Zabs_nat_Z_of_nat.
omega.
field; auto with real.
apply Rle_ge.
unfold Rdiv; repeat apply Rmult_le_pos; try apply sqrt_pos.
left; apply Rinv_0_lt_compat.
now apply sqrt_lt_R0.
now left.
unfold Rdiv; repeat apply Rmult_le_pos; try apply sqrt_pos.
left; apply Rinv_0_lt_compat.
now apply sqrt_lt_R0.
now left.
(* + *)
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
apply sqrt_pos.
rewrite Rmult_comm with (r2:=snd dX).
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
unfold Rdiv; apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
now apply sqrt_lt_R0.
apply Rle_trans with (IZR (ni (fst dX)) * snd dX).
apply Rmult_le_compat_r.
now left.
rewrite <- S_INR, INR_IZR_INZ.
apply IZR_le.
replace (n-0)%nat with n;auto with zarith.
apply Rle_trans with ( ((xmax - xmin) / (fst dX)) * snd dX).
apply Rmult_le_compat_r.
now left.
unfold R_n.ni, jx.
apply (base_Int_part ((xmax - xmin) / fst dX)).
unfold Rdiv; repeat rewrite Rmult_assoc.
apply Rmult_le_compat_l.
apply Rplus_le_reg_l with xmin; ring_simplify.
left; apply xminLtxmax.
apply Rmult_le_reg_l with c; trivial.
apply Rle_trans with (1-eps);[idtac|right; field; auto with real].
apply Rle_trans with (2:=CFL_eps).
unfold CFL, Rdiv; right; ring.
Qed.


Variables M alpha:R.
Hypothesis DL_u0: forall x,
   xmin <= x <= xmax -> exists D1:R,
   forall dx, Rabs dx <= alpha ->
   Rabs ( u0 (x+dx) - u0 x - D1* dx ) <= M * dx*dx.

Lemma unh_bounded_aux0: 0 < fst dX -> 0 < snd dX 
   -> fst dX <= alpha
   -> u0 xmin=0 -> u0 xmax=0 ->  ni_exact xmin xmax dX
   ->  forall i, (0 < i < ni (fst dX) )%Z ->
      Rabs (Ah c (fun j : Z => unh dX j 0) (fst dX) i) <= 2*M*c*c.
intros.
(* *)
assert (K1:(0 < c*c)%R).
now apply Rmult_lt_0_compat.
assert (K2:(0 < fst dX*fst dX)%R).
now apply Rmult_lt_0_compat.
assert (K3:(0 < /(fst dX*fst dX))%R).
auto with real.
apply Rle_trans with ((c*c)*(2*M*fst dX * fst dX)*(/(fst dX*fst dX))).
2: right; field; auto with real.
unfold Ah, Rdiv.
rewrite Rabs_mult, Rabs_mult, Rabs_Ropp.
rewrite Rabs_right, (Rabs_right ((/ (fst dX * fst dX)))).
2: apply Rle_ge; auto with real.
2: apply Rle_ge; auto with real.
apply Rmult_le_compat_r; auto with real.
apply Rmult_le_compat_l; auto with real.
(* *)
assert (forall j, (0 <= j <= ni (fst dX))%Z -> unh dX j 0 = u0h dX j).
assert (forall j, (0 <= j <= ni (fst dX))%Z -> scheme xmin xmax c u0h u1h fh unh dX j 0 = 0).
intros j Hj; apply unh_is_discrete_sol; auto with zarith.
intros j Hj; generalize (H6 j Hj); unfold scheme, Lh.
case (Z_lt_le_dec 0 j).
case (Z_lt_le_dec j (ni (fst dX))).
intros; auto with real.
intros J1 J2 J3; rewrite J3.
replace j with (ni (fst dX)); auto with zarith.
unfold u0h; simpl.
replace (xmin + IZR (ni (fst dX)) * fst dX) with xmax; auto with real.
rewrite H4; field; auto with real.
intros J1 J3; rewrite J3.
replace j with 0%Z; auto with zarith.
unfold u0h; simpl.
replace (xmin + 0 * fst dX) with xmin; auto with real.
simpl; ring.
(* *)
rewrite 3!H6; auto with zarith.
unfold u0h; simpl.
destruct (DL_u0 (xmin + IZR i * fst dX)) as (D1,HD1).
replace (xmin + IZR i * fst dX) with (fst (Xjk xmin dX i 0)) by reflexivity. 
apply Xjk_Betw_j; auto with zarith.
replace (u0 (xmin + IZR (i + 1) * fst dX) - 2 * u0 (xmin + IZR i * fst dX) +
   u0 (xmin + IZR (i - 1) * fst dX)) with
   ((u0 ((xmin + IZR i * fst dX)+fst dX) - u0 (xmin + IZR i * fst dX) -D1*fst dX) 
   +(u0 ((xmin + IZR i * fst dX)+-fst dX) - u0 (xmin + IZR i * fst dX) -D1*(-fst dX))).
apply Rle_trans with (1:=Rabs_triang _ _).
apply Rle_trans with (M*fst dX*fst dX+M*(-fst dX)*(-fst dX))%R.
apply Rplus_le_compat; apply HD1.
rewrite Rabs_right; try apply Rle_ge; auto with real.
rewrite Rabs_Ropp, Rabs_right; try apply Rle_ge; auto with real.
right; ring.
replace (xmin+IZR (i + 1) * fst dX) with (xmin+IZR i * fst dX + fst dX).
replace (xmin+IZR (i - 1) * fst dX) with (xmin+IZR i * fst dX +- fst dX).
ring.
unfold Zminus; rewrite plus_IZR; simpl; ring.
rewrite plus_IZR; simpl; ring.
Qed.


(*
Lemma unh_bounded_aux0: forall r,  0 < fst dX -> 0 < snd dX -> (forall j, Rabs (u0h dX j) <= r) 
   ->  forall i, (0 < i < ni (fst dX) )%Z ->
      Rabs (Ah c (fun j : Z => unh dX j 0) (fst dX) i) <=
          c * (c * (2 * (2 * r))) / (Rsqr (fst dX)). *)



Lemma unh_bounded_aux1: forall r,  0 < fst dX -> 0 < snd dX
   -> fst dX <= alpha
   -> u0 xmin=0 -> u0 xmax=0 ->  ni_exact xmin xmax dX
   -> (forall j, Rabs (u0h dX j) <= r) 
   -> (forall j, u1h dX j =0)
   ->  forall i, (0 <= i <= ni (fst dX) )%Z ->
       Rabs (unh dX i 1%nat) <= r + Rsqr (c*snd dX)*M.
intros.
(* . *)
assert (forall j, (0 <= j <= ni (fst dX))%Z -> Rabs (unh dX j 0) <= r).
intros; assert (scheme xmin xmax c u0h u1h fh unh dX j 0 = 0).
apply unh_is_discrete_sol; auto with zarith.
generalize H9; unfold scheme, Lh.
case (Z_lt_le_dec 0 j).
case (Z_lt_le_dec j (ni (fst dX))).
intros; auto with real.
replace (unh dX j 0) with (u0h dX j); auto with real.
intros V1 V2 V3; rewrite V3, Rabs_R0.
apply Rle_trans with (2:=H5 0%Z); apply Rabs_pos.
intros V1 V3; rewrite V3, Rabs_R0.
apply Rle_trans with (2:=H5 0%Z); apply Rabs_pos.
(* . *)
assert (scheme xmin xmax c u0h u1h fh unh dX i 1 = 0).
apply unh_is_discrete_sol; auto with zarith.
generalize H9; unfold scheme.
case (Z_lt_le_dec 0 i); intros Hi1.
case (Z_lt_le_dec i (ni (fst dX))); intros Hi2.
rewrite H6; unfold Lh.
intros.
replace (unh dX i 1) with 
  (unh dX i 0- snd dX * snd dX / 2 * Ah c (fun j' : Z => unh dX j' 0) (fst dX) i).
apply Rle_trans with (1:=Rabs_triang _ _).
apply Rplus_le_compat.
apply H8; auto.
rewrite Rabs_Ropp.
unfold Rdiv; rewrite Rabs_mult.
rewrite Rabs_right.
apply Rle_trans with (snd dX * snd dX * / 2 * (2*M*c*c)).
2: right; unfold Rsqr; field; auto with real.
repeat rewrite Rmult_assoc.
apply Rmult_le_compat_l; auto with real.
repeat rewrite Rmult_assoc.
repeat apply Rmult_le_compat_l; auto with real.
repeat rewrite <- Rmult_assoc.
apply unh_bounded_aux0; auto.
apply Rle_ge; apply Rle_trans with (snd dX*0); auto with real.
rewrite Rmult_assoc; apply Rmult_le_compat_l; auto with real.
apply Rmult_le_pos; auto with real.
apply Rmult_eq_reg_l with (/snd dX).
apply trans_eq with (-( (unh dX i 1 - unh dX i 0) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => unh dX j' 0) (fst dX) i - 0) + / snd dX * unh dX i 1)%R.
unfold Rdiv; field; auto with real.
rewrite H10; ring.
auto with real.
intros H10; rewrite H10, Rabs_R0.
apply Rle_trans with (0+0); auto with real; apply Rplus_le_compat.
apply Rle_trans with (2:=H5 0%Z); apply Rabs_pos.
repeat apply Rmult_le_pos; auto with real.
apply Rmult_le_reg_l with (fst dX*fst dX).
apply Rmult_lt_0_compat; auto.
apply Rle_trans with (M*fst dX * fst dX);[idtac|right; ring].
destruct (DL_u0 xmin); auto with real.
apply Rle_trans with (Rabs (u0 (xmin + fst dX) - u0 xmin - x * fst dX)).
ring_simplify; apply Rabs_pos.
apply H11.
rewrite Rabs_right; try apply Rle_ge; auto with real.
intros H10; rewrite H10, Rabs_R0.
apply Rle_trans with (0+0); auto with real; apply Rplus_le_compat.
apply Rle_trans with (2:=H5 0%Z); apply Rabs_pos.
repeat apply Rmult_le_pos; auto with real.
apply Rmult_le_reg_l with (fst dX*fst dX).
apply Rmult_lt_0_compat; auto.
apply Rle_trans with (M*fst dX * fst dX);[idtac|right; ring].
destruct (DL_u0 xmin); auto with real.
apply Rle_trans with (Rabs (u0 (xmin + fst dX) - u0 xmin - x * fst dX)).
ring_simplify; apply Rabs_pos.
apply H11.
rewrite Rabs_right; try apply Rle_ge; auto with real.
Qed.




Lemma unh_bounded_aux2: forall r,  0 < fst dX -> 0 < snd dX 
   -> fst dX <= alpha
   -> u0 xmin=0 -> u0 xmax=0 ->  ni_exact xmin xmax dX
   -> (forall j, Rabs (u0h dX j) <= r) 
   -> (forall j, u1h dX j =0)
   -> energy c (unh dX) dX 0%nat  <= 
         (1/2*((xmax-xmin)*Rsqr (snd dX*M*c*c))
                        +1/2*(((xmax - xmin) * (2*M*c*c)) *
              (r + Rsqr (c * snd dX) * M))).
intros.
unfold one_d_wave_equation.energy.
assert (0 <= 1/2)%R.
unfold Rdiv; rewrite Rmult_1_l; auto with real.
assert (0 <= r)%R.
apply Rle_trans with (2:=H5 0%Z); apply Rabs_pos.
assert (0 <= M)%R.
apply Rmult_le_reg_l with (fst dX*fst dX).
apply Rmult_lt_0_compat; auto.
apply Rle_trans with (M*fst dX * fst dX);[idtac|right; ring].
destruct (DL_u0 xmin); auto with real.
apply Rle_trans with (Rabs (u0 (xmin + fst dX) - u0 xmin - x * fst dX)).
ring_simplify; apply Rabs_pos.
apply H9.
rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rplus_le_compat; apply Rmult_le_compat_l; auto with real.
(* . *)
rewrite finite_squared_norm_dx_homogen.
rewrite finite_squared_norm_dx_eq with 
  (f2:= (scalar_multiply_n (snd dX * snd dX / 2) 
     (fun i => if (Z_lt_le_dec 0 i) then
       - Ah c (fun j' : Z => unh dX j' 0) (fst dX) i
       else 0))); auto.
rewrite finite_squared_norm_dx_homogen.
apply Rle_trans with (snd dX * snd dX */ 2*/2 *
  R_n.finite_squared_norm_dx xmin xmax (fst dX) 
   (fun i0 : Z =>
    if Z_lt_le_dec 0 i0 
    then - Ah c (fun j' : Z => unh dX j' 0) (fst dX) i0
    else 0)).
right; field; auto with real.
apply Rle_trans with (snd dX * snd dX */ 2 */2 *((xmax-xmin)*
 Rsqr (2*M*c*c))).
apply Rmult_le_compat_l; auto with real.
repeat apply Rmult_le_pos; auto with real.
apply finite_squared_norm_dx_majoration; auto.
apply Rle_0_sqr.
intros; case (Z_lt_le_dec 0 j); intros.
apply Rsqr_le_abs_1.
rewrite Rabs_Ropp.
rewrite (Rabs_right (2*M*c*c)).
apply unh_bounded_aux0; auto with zarith.
apply Rle_ge; repeat apply Rmult_le_pos; auto with real.
rewrite Rsqr_0; apply Rle_0_sqr.
right; unfold Rsqr; field; auto with real.
intros; unfold subtract_n, scalar_multiply_n.
assert (scheme xmin xmax c u0h u1h fh unh dX i 1 = 0).
apply unh_is_discrete_sol; auto with zarith.
generalize H11; unfold scheme.
case (Z_lt_le_dec 0 i); intros Hi1.
case (Z_lt_le_dec i (ni (fst dX))); intros Hi2.
rewrite H6; unfold Lh.
intros H12.
apply Rmult_eq_reg_l with (/snd dX); auto with real.
apply trans_eq with (-( (unh dX i 1 - unh dX i 0) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => unh dX j' 0) (fst dX) i - 0) + / snd dX * (unh dX i (0+1) - unh dX i 0))%R.
rewrite H12; ring.
rewrite plus_0_l.
unfold Rdiv; field; auto with real.
contradict Hi2; auto with zarith.
replace i with 0%Z; auto with zarith.
intros H12; rewrite plus_0_l, H12.
assert (scheme xmin xmax c u0h u1h fh unh dX 0 0 = 0).
apply unh_is_discrete_sol; auto with zarith.
generalize H13; unfold scheme.
case (Z_lt_le_dec 0 0).
intros V; contradict V; auto with zarith.
intros V1 V2; rewrite V2; ring.
(* . *)
rewrite plus_0_l.
apply Rle_trans with (1:=RRle_abs _).
apply finite_dot_product_le2; auto.
repeat apply Rmult_le_pos; auto with real.
apply Rle_trans with (0+0); auto with real; apply Rplus_le_compat; auto.
repeat apply Rmult_le_pos; auto with real.
intros.
case (Zle_lt_or_eq 0 i); auto with zarith.
intros; left; apply unh_bounded_aux0; auto with zarith.
intros H11; rewrite <- H11.
right.
assert (scheme xmin xmax c u0h u1h fh unh dX 0 1 = 0).
apply unh_is_discrete_sol; auto with zarith.
generalize H12; unfold scheme.
case (Z_lt_le_dec 0 0); intros Hi1; auto.
intros; left.
apply unh_bounded_aux1; auto with zarith.
Qed.



Lemma unh_bounded_aux3: forall r,  0 < fst dX -> 0 < snd dX 
   -> fst dX <= alpha
   -> u0 xmin=0 -> u0 xmax=0 ->  ni_exact xmin xmax dX
   -> (forall j, Rabs (u0h dX j) <= r) 
   -> (forall j, u1h dX j =0)
   -> energy c (unh dX) dX 0%nat  <= 
        (xmax-xmin)*(M*Rsqr c) * (r + 3/2*M*Rsqr (snd dX*c)).
intros.
apply Rle_trans with  (1 / 2 * ((xmax - xmin) * (snd dX * M * c * c)²) +
       1 / 2 * ((xmax - xmin) * (2 * M * c * c) * (r + (c * snd dX)² * M)))%R.
apply unh_bounded_aux2; auto.
unfold Rsqr; right; field.
Qed.


Lemma unh_bounded_prog: forall n r L, (Z_of_nat n <= ni (fst dX))%Z 
   -> 0 < fst dX -> 0 < snd dX 
   -> fst dX <= alpha
   -> u0 xmin=0 -> u0 xmax=0 ->  ni_exact xmin xmax dX
   -> (forall j, Rabs (u0h dX j) <= r) 
   -> (forall j, u1h dX j =0)
   -> (forall j k, fh dX j k =0) ->
       forall j, (0 < j < ni (fst dX) )%Z -> Rabs (unh dX j n) <= L.
intros n r L Hn H0 H1 H2 H3 H4 Hni Hu0 Hu1 Hf j Hj.
apply Rle_trans with (L/sqrt (fst dX))%R.
apply finite_norm_le_reverse with (xmin:=xmin) (xmax:=xmax) (uh:=fun j => unh dX j n); auto with zarith.
apply Rle_trans with (finite_norm_dx (fst dX) (u0h dX) + 
    ((xmax-xmin)/c * (1-eps)) * (sqrt 2 / sqrt (2 * eps - Rsqr eps)) * sqrt (energy c (unh dX) dX 0%nat)).
apply unh_bounded; auto.
unfold u0h; simpl.
rewrite Rmult_0_l, Rplus_0_r; auto with real.
apply Rle_trans with (sqrt (xmax -xmin)*r+L).
apply Rplus_le_compat.
apply finite_norm_le; auto.
left; auto with zarith.
apply Rle_trans with ((xmax - xmin) / c * (1 - eps) * (sqrt 2 / sqrt (2 * eps - eps²)) *
   sqrt  ((xmax-xmin)*(M*Rsqr c) * (r + 3/2*M*Rsqr (snd dX*c)))).
apply Rmult_le_compat_l.

admit.

apply sqrt_le_1.
apply energy_pos with u0h u1h fh eps; auto.

admit.

apply unh_bounded_aux3; auto.

admit.

auto with real.


(
 sqrt (xmax - xmin) * r +
  ((xmax - xmin)  * (1 / sqrt (eps )) *
   sqrt ((xmax - xmin) * M  * (r + 3 / 2 * M * (snd dX * c)²))
  )
)
/ sqrt (fst dX)

(* si xmax-xmin=1, alors
Rabs (unh dX j n) <= 
   (r + /sqrt eps * sqrt M * sqrt (r+3/2*M*(snd dX*c)²))/ fst dX

pour r=1 eps=0.01 et M = 100 
  (1+ 100*sqrt 1.01)/fst dX
*)

Admitted.

End StabFP.
*)


